"""
segmentor classes that takes a streams of dataframes and generate chunks in various ways.

Author: Qu Tang

Date: 02/04/2020

License: GNU v3
"""
import queue
import pandas as pd
from . import extensions as ext
from . import moment
from loguru import logger
from . import operator


class Segmentor(operator.Operator):
    """Base class for segmentors.

    Segmentors are used to segment streaming data and generate chunks in different ways.
    """

    def __init__(self, ref_st: "str, datetime, numpy.datetime64, pandas.Timestamp" = None, st_col: int = 0, et_col: int = None):
        """Create Segmentor instance.

        Arguments:
            ref_st: The reference start time for the first segmented window of data.
            st_col: The column with start time timestamps in the streaming data.
            et_col: The column with stop time timestamps in the streaming data. If it is `None`, `et_col = st_col`.
        """
        super().__init__()
        self._st_col = st_col
        self._et_col = et_col or self._st_col
        self._ref_st = ref_st
        self.reset()

    def set_ref_time(self, ts: "str, datetime, numpy.datetime64, pandas.Timestamp"):
        """Set reference start time.

        Arguments:
            ts: The timestamp to be set.
        """
        self._ref_st = ts

    def reset(self):
        """Reset the segmentor.
        """
        pass

    def run(self, values=None, src=None, context={}):
        self._context = {**self._context, **context}
        for result, new_context in self.segment(values):
            if self._stop:
                break
            self._result.put((result, new_context))

    def segment(self, data: "pandas.Dataframe"):
        """A python generator function to output segmented data.

        The default behavior is to output each row of the burst of streaming data.

        Arguments:
            data: the input burst of streaming data.
        """
        if data is None:
            return
        for index, row in data.iterrows():
            if self._stop:
                break
            yield row, self._context


class SlidingWindowSegmentor(Segmentor):
    """Segment straming data with sliding window method.
    """

    def __init__(self, window_size: float, **kwargs):
        """Create SlidingWindowSegmentor instance.

        Arguments:
            window_size: The window size in seconds.

        Raises:
            ValueError: Raise when window size is smaller than zero.
        """
        super().__init__(**kwargs)
        if window_size == 0:
            raise ValueError('Window size should be greater than zero.')
        self._ws = window_size

    def stop(self):
        super().stop()
        self.reset()

    def reset(self):
        self._current_segment = []
        self._current_seg_st = None
        self._current_seg_et = None
        self._previous_seg_st = None
        self._previous_seg_et = None

    def segment(self, data: 'pandas.Dataframe'):
        """A python generator function to output segmented data.

        It will segment the incoming streaming data with sliding window method and yield segmented data.

        Arguments:
            data: the input burst of streaming data.
        """

        if data is None:
            yield data, self._context
        else:
            et = data.iloc[-1, self._et_col]
            if self._ref_st is not None and moment.Moment(self._ref_st).to_unix_timestamp() > moment.Moment(et).to_unix_timestamp():
                logger.warning(
                    'Referenced start time is after the end time of the input data, this generates no segments from the data.')
                return
            segments = self._extract_segments(data)
            for segment, seg_st, seg_et in segments:
                if self._stop:
                    break
                self._current_seg_st = self._current_seg_st or seg_st
                self._current_seg_et = self._current_seg_et or seg_et
                if self._current_seg_st == seg_st and self._current_seg_et == seg_et:
                    self._current_segment.append(segment)
                else:
                    self._current_segment = pd.concat(
                        self._current_segment, axis=0, sort=False, ignore_index=True)
                    result = self._current_segment
                    new_context = {
                        "start_time": self._current_seg_st,
                        "stop_time": self._current_seg_et,
                        "prev_start_time": self._previous_seg_st,
                        "prev_stop_time": self._previous_seg_et,
                    }
                    self._context = {**self._context, **new_context}
                    self._current_segment = [segment]
                    self._previous_seg_st = self._current_seg_st
                    self._previous_seg_et = self._current_seg_et
                    self._current_seg_st = seg_st
                    self._current_seg_et = seg_et
                    yield result, self._context

    def _extract_segments(self, data):
        if data.empty:
            return []
        data_et = data.iloc[-1, self._et_col]
        data_st = data.iloc[0, self._st_col]
        if self._ref_st is None:
            self._ref_st = data_st
        window_ts_marks = pd.date_range(start=self._ref_st,
                                        end=data_et,
                                        freq=str(self._ws * 1000) + 'ms'
                                        )
        self._ref_st = window_ts_marks[-1]
        segments = []
        for seg_st in window_ts_marks:
            seg_et = seg_st + \
                pd.Timedelta(self._ws * 1000, unit='ms')
            segment = ext.pandas.segment_by_time(
                data,
                seg_st=seg_st,
                seg_et=seg_et,
                st_col=self._st_col,
                et_col=self._et_col
            )
            segments.append((segment, seg_st, seg_et))
        return segments





from ..extensions.sklearn import ELM
from sklearn import datasets
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.utils.estimator_checks import check_estimator
import numpy as np


def test_elm():
    bunch = datasets.load_iris(return_X_y=False)
    X = bunch['data']
    y = np.array([bunch['target_names'][i] for i in bunch['target']])
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=None)
    print(X_train.shape)
    print(X_test.shape)
    print(bunch['target'][:5])
    print(y[:5])
    print(bunch['target_names'])

    elm = ELM(hidden_layer_sizes=(10000,))
    elm.fit(X_train, y_train, labels=bunch['target_names'])
    pred_y = elm.predict(X_test)
    pred_probs = elm.predict_proba(X_test)
    print(accuracy_score(y_test, pred_y))
    print(pred_y[:5])
    print(bunch['target_names'])
    print(pred_probs[:5, :])
    check_estimator(elm)
    assert True



import glob
import os
from . import constants
from . import helper
import pandas as pd
import datetime as dt
from loguru import logger


def get_processed_path(dataset_path):
    return os.path.join(
        dataset_path, constants.PROCESSED_FOLDER)


def get_processed_files(dataset_path):
    processed_files = glob.glob(os.path.join(
        get_processed_path(dataset_path), '*.csv'))
    results = {}
    for f in processed_files:
        results[os.path.basename(f).split('.')[0]] = f
    return results


def get_location_mappings(dataset_path):
    filepath = os.path.join(
        dataset_path, constants.META_FOLDER, constants.META_LOCATION_MAPPING_FILENAME)
    if not os.path.exists(filepath):
        return None
    mappings = pd.read_csv(filepath, header=0)
    return mappings


def get_sensor_placement(dataset_path, pid, sid):
    mapping = get_location_mappings(dataset_path)
    condition = (mapping.iloc[:, 0] == pid) & (
        mapping.iloc[:, 1] == sid)
    placement = mapping.loc[condition, :].iloc[:, 2].values[0]
    return placement.lower()


def get_subjects_info(dataset_path):
    filepath = os.path.join(
        dataset_path, constants.META_FOLDER, constants.META_SUBJECTS_FILENAME)
    if not os.path.exists(filepath):
        return None
    subjects = pd.read_csv(filepath, header=0)
    return subjects


def get_class_category(dataset_path):
    filepath = os.path.join(
        dataset_path, constants.META_FOLDER, 'muss_class_labels.csv')
    if not os.path.exists(filepath):
        return None
    class_category = pd.read_csv(filepath, header=0)
    return class_category


def get_offset_mappings(dataset_path):
    filepath = os.path.join(
        dataset_path, constants.META_FOLDER, 'offset_mappings.csv')
    if not os.path.exists(filepath):
        return None
    offset_mappings = pd.read_csv(filepath, header=0)
    return offset_mappings


def get_orientation_corrections(dataset_path):
    filepath = os.path.join(
        dataset_path, constants.META_FOLDER, 'orientation_corrections.csv')
    if not os.path.exists(filepath):
        return None
    orientation_corrections = pd.read_csv(filepath, header=0)
    return orientation_corrections


def get_pids(dataset_path):
    pids = filter(lambda name: name not in [
        constants.PROCESSED_FOLDER, constants.META_FOLDER, '.git', '.gitignore'], os.listdir(dataset_path))
    return list(sorted(pids))


def get_subject_log(dataset_path, pid, log_filename=None):
    if log_filename is None:
        return os.path.join(dataset_path, pid, constants.SUBJECT_LOG_FOLDER)
    else:
        return os.path.join(dataset_path, pid, constants.SUBJECT_LOG_FOLDER, log_filename)


def get_subject_meta(dataset_path, pid, meta_filename=None):
    if meta_filename is None:
        return os.path.join(dataset_path, pid, constants.SUBJECT_META_FOLDER)
    else:
        return os.path.join(dataset_path, pid, constants.SUBJECT_META_FOLDER, meta_filename)


def get_sensor_files(pid, dataset_path, sid="", given_date=None, data_type=""):
    if given_date is not None:
        date_hour_str = given_date.strftime(f'%Y-%m-%d{os.sep}%H')
    name_pattern = f'*{data_type}*{sid}*.sensor.csv*'
    path_pattern = '**' if given_date is None else f'{date_hour_str}'
    files = glob.glob(os.path.join(dataset_path, pid,
                                   constants.MASTER_FOLDER, path_pattern, name_pattern), recursive=True)
    return sorted(files)


def get_annotation_files(pid, dataset_path, annotation_type=None, annotator=None):
    annotation_type = annotation_type or ""
    annotator = annotator or ""
    files = glob.glob(os.path.join(dataset_path, pid,
                                   constants.MASTER_FOLDER, '**', '*{}*{}*.annotation.csv*'.format(annotation_type, annotator)), recursive=True)
    return sorted(files)


def get_date_folders(pid, dataset_path):
    folder_names = os.listdir(os.path.join(
        dataset_path, pid, constants.MASTER_FOLDER))
    if '-' in folder_names[0]:
        sep = '-'
    else:
        sep = os.sep
    if sep == os.sep:
        date_folders = glob.glob(os.path.join(
            dataset_path, pid, constants.MASTER_FOLDER, '*', '*', '*', '*'), recursive=True)
    elif sep == '-':
        date_folders = glob.glob(os.path.join(
            dataset_path, pid, constants.MASTER_FOLDER, '*', '*'), recursive=True)
    date_folders = list(filter(lambda path: os.path.isdir(path), date_folders))
    return date_folders


def get_session_span(pid, dataset_path):
    date_folders = get_date_folders(pid, dataset_path)
    folders_as_ts = list(
        map(helper.parse_date_from_filepath, date_folders))
    folders_as_ts = sorted(folders_as_ts)
    return folders_as_ts[0], folders_as_ts[-1] + dt.timedelta(hours=1)


def get_session_start_time(pid, dataset_path, round_to='hour'):
    smallest = dt.datetime.now()
    filepaths = get_sensor_files(
        pid, dataset_path) + get_annotation_files(pid, dataset_path)
    for path in filepaths:
        timestamp = helper.parse_timestamp_from_filepath(path, ignore_tz=True)
        if timestamp < smallest:
            smallest = timestamp
    if round_to == 'minute':
        smallest = smallest.replace(microsecond=0, second=0)
    elif round_to == 'second':
        smallest = smallest.replace(microsecond=0)
    else:
        smallest = smallest.replace(microsecond=0, second=0, minute=0)
    return smallest


def traverse_dataset(dataset_path):
    """Traverse the given raw mhealth dataset to parse meta info, sensor and annotation files for each participant.

    Args:
        dataset_path (str): The filepath to the raw dataset.

    Returns:
        dict: python dict object storing the traversed paths and meta info for the given dataset.

        {
            "meta": {
                "location_mapping": pandas.DataFrame, location_mappings,
                "name": str, dataset name,
                "root": str, dataset_path,
                "subjects": pandas.DataFrame, subjects' geographic info
            },
            "processed": {
                "muss": "muss.csv",
                ...
            },
            "subjects": {
                "pid_1": {
                    "sensors": {
                        "DW": [
                            "/path/to/file1.sensor.csv",
                            ...
                        ],
                        ...
                    },
                    "annotations": {
                        "annotation_type": [
                            "/path/to/file1.annotation.csv",
                            ...
                        ],
                        ...
                    }
                }
            }
        }
    """
    logger.warning(
        'This method has been deprecated. And will be removed in version 1.2.0.')

    def _parse_placements(placements):
        return [helper.parse_placement_from_str(p) for p in placements]

    def _get_placements_from_location_mappings(pid, location_mappings):
        filter_pid = location_mappings[constants.FEATURE_SET_PID_COL] == pid
        return location_mappings.loc[filter_pid, 'SENSOR_PLACEMENT'].values.tolist()

    def _get_sids_from_location_mappings(pid, location_mappings):
        filter_pid = location_mappings[constants.FEATURE_SET_PID_COL] == pid
        return location_mappings.loc[filter_pid, 'SENSOR_ID'].values.tolist()

    def _parse_annotation_types(annotation_files):
        return list(set([helper.parse_annotation_type_from_filepath(filepath)
                         for filepath in annotation_files]))

    dataset_dict = {'meta': {}, 'subjects': {}}
    ids = get_pids(dataset_path)

    # Get meta files
    location_mappings = get_location_mappings(
        dataset_path)
    subjects = get_subjects_info(dataset_path)
    class_category = get_class_category(dataset_path)
    offset_mappings = get_offset_mappings(dataset_path)
    orientation_corrections = get_orientation_corrections(dataset_path)
    dataset_dict['meta']['location_mapping'] = location_mappings
    dataset_dict['meta']['subjects'] = subjects
    dataset_dict['meta']['class_category'] = class_category
    dataset_dict['meta']['offset_mapping'] = offset_mappings
    dataset_dict['meta']['orientation_corrections'] = orientation_corrections
    dataset_dict['meta']['root'] = dataset_path
    dataset_dict['meta']['name'] = os.path.basename(dataset_path)

    # Get processed files
    dataset_dict['processed'] = get_processed_files(dataset_path)

    for pid in ids:
        dataset_dict['subjects'][pid] = {'sensors': {}, 'annotations': {}}

        # Get sensor files
        sids = _get_sids_from_location_mappings(pid, location_mappings)
        placements = _get_placements_from_location_mappings(
            pid, location_mappings)
        placements = _parse_placements(placements)

        for sid, p in zip(sids, placements):
            dataset_dict['subjects'][pid]['sensors'][p] = get_sensor_files(
                pid, dataset_path, sid=sid)

        # Get annotation files
        annotation_files = get_annotation_files(pid, dataset_path)
        annotation_types = _parse_annotation_types(annotation_files)
        for annotation_type in annotation_types:
            dataset_dict['subjects'][pid]['annotations'][annotation_type] = get_annotation_files(
                pid, dataset_path, annotation_type=annotation_type)
    return dataset_dict



from .. import dataset
from .. import env
from .. import segmentor
from .. import spades_lab as slab
from .. import feature_vector as fv
from .. import cli
import os
import shutil
import pandas as pd
from dataclasses import asdict
import pytest
from loguru import logger


class TestSpadesLab:
    def test_get_dataset_path(self):
        dataset_path = dataset.get_dataset_path('spades_lab')
        assert dataset_path == os.path.join(
            env.get_data_home(), 'spades_lab')
        assert os.path.exists(dataset_path)

    def test_load_dataset(self):
        dataset_dict = dataset.load_dataset('spades_lab')
        assert type(dataset_dict) == dict
        assert 'subjects' in dataset_dict.keys()
        assert 'processed' in dataset_dict.keys()


class TestSensorObj:
    @pytest.mark.parametrize("seg", [None, 'sliding_window'])
    @pytest.mark.parametrize("placement_parser", [None, slab.get_sensor_placement])
    @pytest.mark.parametrize("stream_id", [None, "test"])
    def test_get_stream(self, seg, placement_parser, stream_id):
        dataset_path = dataset.get_dataset_path('spades_lab')
        mh_ds = dataset.MHDataset(
            path=dataset_path, name='spades_lab', input_type=dataset.InputType.MHEALTH_FORMAT)
        mh_ds.set_placement_parser(placement_parser)
        if seg == 'sliding_window':
            seg = segmentor.SlidingWindowSegmentor(window_size=12)
        stream = mh_ds.subjects[0].sensors[0].get_stream(
            seg=seg, stream_id=stream_id, window_size=12)
        stream.start()
        count = 0
        for data, context in stream.get_result():
            logger.debug(count)
            count += 1
            assert data.shape[0] == 12 * 80
            assert data.shape[1] == 4
            if stream_id is not None:
                assert context['data_id'] == 'test'
            elif placement_parser is None:
                assert context['data_id'] == 'TAS1E23150066-AccelerationCalibrated-stream'
            else:
                assert context['data_id'] == 'DT'
            if count == 5:
                logger.debug('stop')
                stream.stop()

    def test_compute_features_on_the_fly(self):
        dataset_path = dataset.get_dataset_path('spades_lab')
        mh_ds = dataset.MHDataset(
            path=dataset_path, name='spades_lab', input_type=dataset.InputType.MHEALTH_FORMAT)
        mh_ds.set_placement_parser(slab.get_sensor_placement)
        stream = mh_ds.subjects[0].sensors[0].get_stream(
            seg=None, stream_id=None, window_size=12.8)
        stream.start()
        count = 0
        for data, context in stream.get_result():
            logger.debug(count)
            count += 1
            fv_df, fv_names = fv.inertial.single_triaxial(
                data, sr=80, st=context['start_time'], et=context['stop_time'], selected=['MEAN'])
            assert fv_df.iloc[0, 1].timestamp(
            ) == context['start_time'].timestamp()
            assert fv_df.iloc[0, 2].timestamp(
            ) == context['stop_time'].timestamp()
            assert len(fv_names) == 1
            assert fv_df.shape[0] == 1
            if count == 5:
                logger.debug('stop')
                stream.stop()


class TestMHDataset:
    def test_mh_dataset(self):
        dataset_path = dataset.get_dataset_path('spades_lab')
        mh_ds = dataset.MHDataset(
            path=dataset_path, name='spades_lab', input_type=dataset.InputType.MHEALTH_FORMAT)
        assert mh_ds.name == 'spades_lab'
        assert mh_ds.subjects[0].pid == 'SPADES_1'
        assert mh_ds.subjects[0].demography is None
        assert mh_ds.subjects[0].sensors[0].data_type == 'AccelerationCalibrated'
        assert mh_ds.subjects[0].annotations[0].annotation_type == 'SPADESInLab'
        assert len(mh_ds.subjects[0].sensors[0].paths) == 3
        assert mh_ds.subjects[0].sensors[0].start_time.timestamp(
        ) == pd.Timestamp('2015-09-24 14:22:48.013000').timestamp()


class TestSGDataset:
    def test_sg_dataset(self):
        spades_lab_path = dataset.get_dataset_path('spades_lab')
        dataset_path = os.path.join(
            spades_lab_path, 'SPADES_1', 'Derived', 'signaligner')
        if not os.path.exists(dataset_path):
            cli.convert_to_signaligner_both(spades_lab_path, 'SPADES_1', 80)
        sg_ds = dataset.SGDataset(
            path=dataset_path, name='spades_1_sg', sensor_folder='SPADES_1_2015092414_2015092417_sensors', label_folder='SPADES_1_2015092414_2015092417_labelsets')
        assert sg_ds.name == 'spades_1_sg'
        assert sg_ds.subjects[0].pid == 'spades_1_sg'
        assert sg_ds.subjects[0].demography is None
        assert sg_ds.subjects[0].sensors[0].data_type == 'SPADES_1_ActigraphGT9X_TAS1E23150066-AccelerationCalibrated_2015092414_2015092417.sensor'
        assert sg_ds.subjects[0].sensors[0].sid == 'TAS1E23150066'
        assert sg_ds.subjects[0].sensors[0].sr == 80
        assert sg_ds.subjects[0].sensors[0].grange == 8
        assert sg_ds.subjects[0].annotations[0].annotation_type == 'SPADES_1_diego-SPADESInLab_2015092414_2015092417.annotation'
        assert len(sg_ds.subjects[0].sensors[0].paths) == 1



from pathos import pools
import os
import dill
import datetime as dt


class _AppState:
    def __init__(self):
        self.origin_dataset = []
        self.origin_labels = []
        self.origin_label_candidates = []
        self.origin_model = None
        self.origin_validate_results = None

        self.new_activity_name = ""
        self.data_collection_labels = []
        self.data_collection_label_candidates = []
        self.new_dataset = None

        self.new_labels = []
        self.new_model = None
        self.new_validate_results = None

        self.nearby_devices = None

        self.test_stack = []

        self.output_folder = AppState._path
        self.pid = 'ARUS_' + dt.datetime.now().strftime('%Y%m%d%H%M%S')

        self.task_pool = pools.ProcessPool(nodes=6)
        self.task_pool.close()
        self.io_pool = pools.ThreadPool(nodes=4)
        self.io_pool.close()


class AppState:
    def __init__(self):
        pass

    _instance = None

    _path = os.path.join(
        os.path.expanduser("~"), 'arus')

    _snapshot_path = os.path.join(
        _path, 'snapshots')

    @staticmethod
    def getInstance():
        if AppState._instance is None:
            AppState._instance = _AppState()
        return AppState._instance

    @staticmethod
    def snapshot():
        output_path = os.path.join(
            AppState._snapshot_path, AppState._instance.pid + '.pkl')
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, 'wb') as f:
            dill.dump(AppState._instance, f)

    @staticmethod
    def restore(file_path):
        input_path = file_path
        with open(input_path, 'rb') as f:
            AppState._instance = dill.load(f)
        return True

    @staticmethod
    def reset():
        AppState._instance = _AppState()



from . import constants
import datetime as dt
import pandas as pd
import os
import re
from .. import moment


class ParseError(Exception):
    pass


def parse_column_names_from_data_type(data_type):
    if data_type in ['AccelerometerCalibrated', 'IMUAccelerometerCalibrated', 'AccelerationCalibrated']:
        return ['ACCELEROMETER_X', 'ACCELEROMETER_Y', 'ACCELEROMETER_Z']
    elif data_type in ['IMUTenAxes']:
        return ["ACCELEROMETER_X", "ACCELEROMETER_Y", "ACCELEROMETER_Z", "TEMPERATURE", "GYROSCOPE_X", "GYROSCOPE_Y", "GYROSCOPE_Z", "MAGNETOMETER_X", "MAGNETOMETER_Y", "MAGNETOMETER_Z"]
    elif data_type in ['IMUGyroscope']:
        return ["GYROSCOPE_X", "GYROSCOPE_Y", "GYROSCOPE_Z"]
    elif data_type in ['IMUMagnetometer']:
        return ["MAGNETOMETER_X", "MAGNETOMETER_Y", "MAGNETOMETER_Z"]
    elif data_type in ['IMUTemperature']:
        return ['TEMPERATURE']
    else:
        raise NotImplementedError(
            f"The given data type {data_type} is not supported")


def parse_column_names_from_filepath(filepath):
    data_type = parse_data_type_from_filepath(filepath)
    return parse_column_names_from_data_type(data_type)


def parse_placement_from_str(placement_str):
    result = ''
    placement_str = placement_str.lower()
    if 'nondominant' in placement_str or 'non-dominant' in placement_str or 'non dominant' in placement_str or placement_str.startswith('nd'):
        result = 'ND'
    elif 'dominant' in placement_str or placement_str.startswith('d'):
        result = 'D'
    if 'ankle' in placement_str or placement_str.endswith('da'):
        result += 'A'
    elif 'wrist' in placement_str or placement_str.endswith('dw'):
        result += 'W'
    elif 'waist' in placement_str or 'hip' in placement_str or placement_str.endswith('dh'):
        result += 'H'
    elif 'thigh' in placement_str or placement_str.endswith('dt'):
        result += 'T'
    return result


def parse_date_from_filepath(filepath, ignore_tz=True):
    tokens = filepath.split(constants.MASTER_FOLDER)[1].split(os.sep)
    if '-' in tokens[1]:
        sep = '-'
    else:
        sep = os.sep
    if sep == os.sep:
        hour = tokens[4].split('-')[0]
        day = tokens[3]
        month = tokens[2]
        year = tokens[1]
    elif sep == '-':
        hour = tokens[2].split('-')[0]
        sub_tokens = tokens[1].split('-')
        day = sub_tokens[-1]
        month = sub_tokens[1]
        year = sub_tokens[0]
    file_date = dt.datetime(year=int(year), month=int(month), day=int(
        day), hour=int(hour), minute=0, second=0, microsecond=0)
    return file_date


def parse_timestamp_from_filepath_content(filepath, ts_col=0, ignore_tz=True):
    first_row = pd.read_csv(filepath, nrows=1, header=0, parse_dates=[
        ts_col], infer_datetime_format=True)
    return first_row.iloc[0, ts_col]


def parse_timestamp_from_filepath(filepath, ignore_tz=True):
    filename = os.path.basename(filepath)
    if filename.endswith('gz'):
        timestamp_index = -4
    else:
        timestamp_index = -3
    timestamp_str = filename.split('.')[timestamp_index]
    if ignore_tz:
        timestamp_str = timestamp_str[:-6]
        result = dt.datetime.strptime(
            timestamp_str, constants.FILE_TIMESTAMP_FORMAT)
    else:
        timestamp_str = timestamp_str.replace('P', '+').replace('M', '-')
        result = dt.datetime.strptime(
            timestamp_str, constants.FILE_TIMESTAMP_FORMAT_WITH_TZ)
    return result


def parse_annotation_type_from_filepath(filepath):
    return os.path.basename(filepath).split('.')[0]


def parse_annotator_from_filepath(filepath):
    return os.path.basename(filepath).split('.')[1]


def parse_sensor_type_from_filepath(filepath):
    result = os.path.basename(filepath).split('.')[0].split('-')[0]
    return result


def parse_data_type_from_filepath(filepath):
    return os.path.basename(filepath).split('.')[0].split('-')[1]


def parse_version_code_from_filepath(filepath):
    return os.path.basename(filepath).split('.')[0].split('-')[2]


def parse_sensor_id_from_filepath(filepath):
    return os.path.basename(filepath).split('.')[1].split('-')[0]


def parse_data_id_from_filepath(filepath):
    return os.path.basename(filepath).split('.')[1]


def parse_pid_from_filepath(filepath):
    try:
        assert is_mhealth_filepath(
            filepath) or is_mhealth_flat_filepath(filepath)
        pid = os.path.basename(
            os.path.dirname(
                filepath.split(constants.MASTER_FOLDER)[
                    0].split(constants.DERIVED_FOLDER)[0]
            )
        )
        return pid
    except Exception:
        raise ParseError('Fail to parse pid for the given filepath')


def parse_subject_path_from_filepath(filepath):
    try:
        assert is_mhealth_filepath(
            filepath) or is_mhealth_flat_filepath(filepath)
        subject_folder = os.path.dirname(
            filepath.split(constants.MASTER_FOLDER)[
                0].split(constants.DERIVED_FOLDER)[0]
        )
        return subject_folder
    except Exception:
        raise ParseError('Fail to parse pid for the given filepath')


def parse_filetype_from_filepath(filepath):
    filename = os.path.basename(filepath)
    if filename.endswith('gz'):
        return filename.split('.')[-3]
    else:
        return filename.split('.')[-2]


def parse_datetime_columns_from_filepath(filepath):
    """Utility to get the timestamp column indices given file type

    Args:
        filepath (str): mhealth file path.

    Returns:
        col_indices (list): list of column indices (0 based)
    """
    filetype = parse_filetype_from_filepath(filepath)
    if filetype == constants.SENSOR_FILE_TYPE:
        return [0]
    elif filetype in [constants.ANNOTATION_FILE_TYPE, constants.FEATURE_FILE_TYPE, constants.CLASS_FILE_TYPE, constants.FEATURE_SET_FILE_TYPE]:
        return [0, 1, 2]
    else:
        raise NotImplementedError(
            'The given file type {} is not supported'.format(filetype))


def format_columns(data, filetype):
    data = data.rename(columns={data.columns[0]: constants.TIMESTAMP_COL})
    if filetype == constants.ANNOTATION_FILE_TYPE:
        data.columns = constants.FEATURE_SET_TIMESTAMP_COLS + \
            [constants.ANNOTATION_LABEL_COL]
    elif filetype == constants.SENSOR_FILE_TYPE:
        # COLUMN names should be A-Z0-9_
        data.columns = list(
            map(lambda col: col.upper().replace(' ', '_'), data.columns))
    return data


def format_file_timestamp_from_data(data, filetype):
    if filetype == constants.SENSOR_FILE_TYPE:
        col = 0
    else:
        col = 1
    st = moment.Moment(data.iloc[0, col]).to_datetime(
        tz=moment.Moment.get_local_timezone())
    timestamp_str = st.strftime(constants.FILE_TIMESTAMP_FORMAT)[:-3]
    timestamp_str += '-' + \
        st.strftime('%z').replace('-', 'M').replace('+', 'P')
    return timestamp_str


def format_date_folder_path_from_data(data, filetype):
    if filetype == constants.SENSOR_FILE_TYPE:
        col = 0
    else:
        col = 1
    st = moment.Moment(data.iloc[0, col]).to_datetime(
        tz=moment.Moment.get_local_timezone())
    year = st.strftime('%Y')
    month = st.strftime('%m')
    day = st.strftime('%d')
    hour = st.strftime('%H')
    return year + '-' + month + '-' + day + os.sep + hour


def compare_two_mhealth_filepaths(filepath1, filepath2):
    sections1 = [os.path.dirname(filepath1)]
    sections2 = [os.path.dirname(filepath2)]
    name1 = os.path.basename(filepath1).strip('.gz')
    name2 = os.path.basename(filepath2).strip('.gz')
    sections1 += name1.split('.')
    sections2 += name2.split('.')
    if len(sections1) != len(sections2):
        return False
    for section1, section2 in zip(sections1, sections2):
        if section1 != section2 and sections1.index(section1) != 3:
            return False
    return True


def transform_class_category(input_label, class_category, input_category,  output_category):
    cond = class_category[input_category] == input_label
    return class_category.loc[cond, output_category].values[0]


def is_mhealth_filepath(filepath):
    """Validate if input file path is in mhealth format

    Args:
        filepath (str): input file path

    Returns:
        is_mhealth (bool): `True` if the input is in mhealth format
    """
    filepath = os.path.abspath(filepath)
    matched = re.search(
        constants.MHEALTH_FILEPATH_PATTERN,
        filepath)
    return matched is not None


def is_mhealth_flat_filepath(filepath):
    """Validate if input file path is in mhealth format (flat structure)

    The flat structure stores all files directly in the `MasterSynced` folder in the pid folder, ignoring all date and hour folders.

    Args:
        filepath (str): input file path

    Returns:
        is_mhealth (bool): `True` if the input is in mhealth flat format
    """
    matched = re.search(
        constants.MHEALTH_FLAT_FILEPATH_PATTERN,
        os.path.abspath(filepath)
    )
    return matched is not None


def is_mhealth_filename(filepath):
    filename = os.path.basename(filepath)

    sensor_filename_pattern = r"^{}\-{}\-{}\.{}\-{}\.{}\.sensor\.csv(\.gz)*$".format(
        constants.CAMELCASE_PATTERN, constants.CAMELCASE_PATTERN,
        constants.VERSIONCODE_PATTERN, constants.SID_PATTERN, constants.CAMELCASE_PATTERN, constants.FILE_TIMESTAMP_PATTERN
    )

    annotation_filename_pattern = r"^{}\.{}\-{}\.{}\.annotation\.csv(\.gz)*$".format(
        constants.CAMELCASE_PATTERN, constants.ANNOTATOR_PATTERN, constants.CAMELCASE_PATTERN, constants.FILE_TIMESTAMP_PATTERN)

    sensor_matched = re.search(
        sensor_filename_pattern,
        filename
    )

    annotation_matched = re.search(
        annotation_filename_pattern,
        filename
    )
    return sensor_matched is not None or annotation_matched is not None



from .. import generator
from .. import mhealth_format as mh
import pandas as pd
from pyarrow import csv
import datetime
import os

ACTIGRAPH_TEMPLATE = """------------ Data File Created By ActiGraph GT3X+ ActiLife v6.13.3 Firmware v2.5.0 date format M/d/yyyy at {} Hz  Filter Normal -----------
Serial Number: {}
Start Time {}
Start Date {}
Epoch Period (hh:mm:ss) 00:00:00
Download Time {}
Download Date {}
Current Memory Address: 0
Current Battery Voltage: 4.21     Mode = 12
--------------------------------------------------"""


class ActigraphSensorFileGenerator(generator.Generator):
    def __init__(self, *filepaths, has_ts=True, has_header=True, **kwargs):
        super().__init__(**kwargs)
        self._filepaths = filepaths
        self._reader = None
        self._has_ts = has_ts
        self._has_header = has_header

    def run(self, values=None, src=None, context={}):
        for filepath in self._filepaths:
            self._reader = ActigraphReader(
                filepath, self._has_ts, self._has_header)
            self._reader.read(chunksize=self._buffer_size)
            for data in self._reader.get_data():
                if self._stop:
                    break
                result = self._buffering(data)
                if result is not None:
                    self._result.put((result, self._context))
            if self._stop:
                break
        self._result.put((None, self._context))


class ActigraphReader:
    def __init__(self, filepath, has_ts=True, has_header=True):
        self._filepath = filepath
        self._data = None
        self._iterator = None
        self._meta = None
        self._has_ts = has_ts
        self._has_header = has_header

    def get_data(self):
        assert self._meta is not None
        ts_func = convert_actigraph_imu_timestamp if self._meta[
            'IMU'] else convert_actigraph_timestamp
        if self._data is not None:
            data = self._data.copy()
            if self._has_ts:
                data.iloc[:, 0] = ts_func(data.iloc[:, 0])
            else:
                ts = generate_timestamps(
                    self._meta['START_TIME'], self._meta['SAMPLING_RATE'], data.shape[0])[:-1]
                data.insert(0, 'ts', ts)
            data = mh.helper.format_columns(
                data, filetype=mh.constants.SENSOR_FILE_TYPE)
            yield data
        else:
            st = self._meta['START_TIME']
            for data in self._iterator:
                if self._has_ts:
                    data.iloc[:, 0] = ts_func(data.iloc[:, 0])
                else:
                    ts = generate_timestamps(
                        st, self._meta['SAMPLING_RATE'], data.shape[0])
                    data.insert(0, 'ts', ts[:-1])
                    st = ts[-1]
                data = mh.helper.format_columns(
                    data, filetype=mh.constants.SENSOR_FILE_TYPE)
                yield data

    def get_meta(self):
        return self._meta.copy()

    def read(self, **kwargs):
        self.read_meta()
        self.read_csv(**kwargs)
        return self

    def read_csv(self, chunksize=None):
        if self._has_ts:
            columns = ['HEADER_TIME_STAMP', 'X', 'Y', 'Z']
        else:
            columns = ['X', 'Y', 'Z']
        if self._has_header:
            header = 0
            columns = None
        else:
            header = None
        if chunksize is None:
            read_opts = csv.ReadOptions(skip_rows=10, column_names=columns)
            reader = csv.read_csv(self._filepath, read_options=read_opts)
            reader = reader.to_pandas()
        else:
            reader = pd.read_csv(
                self._filepath, skiprows=10, chunksize=chunksize, header=header, names=columns, engine='c', memory_map=True)
        if type(reader) == pd.DataFrame:
            self._data = reader
        else:
            self._iterator = reader
        return self

    def read_meta(self):
        with open(self._filepath, 'r') as f:
            first_line = f.readline()
            second_line = f.readline()
            third_line = f.readline()
            fourth_line = f.readline()
            is_imu = 'IMU' in first_line
            firmware = list(
                filter(lambda token: token.startswith('v'), first_line.split(" ")))[1]
            sr = int(
                list(filter(lambda token: token.isnumeric(), first_line.split(" ")))[0])
            sid = second_line.split(" ")[-1].strip()
        if 'TAS' in sid and is_imu:
            g_range = 16
        elif 'TAS' in sid and not is_imu:
            g_range = 8
        else:
            g_range = None

        st_str = fourth_line.split(' ')[2] + ' ' + third_line.split(' ')[2]
        st = pd.Timestamp(st_str)

        self._meta = {
            'START_TIME': st,
            'VERSION_CODE': firmware,
            'SAMPLING_RATE': sr,
            'SENSOR_ID': sid,
            'IMU': is_imu,
            'DYNAMIC_RANGE': g_range
        }
        return self


def generate_timestamps(start_time, sr, n):
    return pd.date_range(start=start_time, periods=n + 1, freq=str(1000.0/sr) + 'L')


def convert_actigraph_timestamp(timestamps):
    """Convert elements in the timestamp columns of the input dataframe to `datetime64[ms]` type.

    Args:
        timestamps

    Returns:
        result (same as input)
    """

    result = pd.to_datetime(
        timestamps, format='%m/%d/%Y %H:%M:%S.%f')
    result = result.astype('datetime64[ms]')
    return result


def convert_actigraph_imu_timestamp(timestamps):
    result = pd.to_datetime(timestamps, infer_datetime_format=True)
    result = result.astype('datetime64[ms]')
    return result


def save_as_actigraph(out_df, output_filepath, sid=None, session_st=None, session_et=None, sr=50, **kwargs):
    sid = sid or 'CLE2B2013XXXX'
    session_st = session_st or out_df.iloc[0, 0].to_datetime()
    session_et = session_et or out_df.iloc[-1, 0].to_datetime()
    meta_sdate_str = '{dt.month}/{dt.day}/{dt.year}'.format(
        dt=session_st)
    meta_stime_str = session_st.strftime('%H:%M:%S')
    meta_edate_str = '{dt.month}/{dt.day}/{dt.year}'.format(
        dt=session_et)
    meta_etime_str = (session_et +
                      datetime.timedelta(hours=1)).strftime('%H:%M:%S')
    col_names = out_df.columns
    col_names = list(map(lambda name: _format_column_name(name), col_names))
    if not os.path.exists(output_filepath):
        # create
        with open(output_filepath, mode='w', encoding='utf-8') as f:
            f.write(ACTIGRAPH_TEMPLATE.format(
                sr, sid, meta_stime_str, meta_sdate_str, meta_etime_str, meta_edate_str))
            f.write('\n')
    # append
    out_df.to_csv(output_filepath, **kwargs)


def _format_column_name(name):
    tokens = name.split('_')
    tokens[0] = tokens[0].lower().title()
    return " ".join(tokens)



from ...core.pipeline import Pipeline
from ...core.stream.generator_stream import GeneratorSlidingWindowStream
from ...core.accelerometer import generator
from datetime import datetime
import pandas as pd
import numpy as np
import sys
import pytest


def _pipeline_test_processor(chunk_list, **kwargs):
    import pandas as pd
    result = {'NAME': [],
              'START_TIME': [], 'STOP_TIME': []}
    for data, st, et, prev_st, prev_et, name in chunk_list:
        result['NAME'].append(name)
        result['START_TIME'].append(data.iloc[0, 0])
        result['STOP_TIME'].append(data.iloc[-1, 0])
    result = pd.DataFrame.from_dict(result)
    return result


@pytest.mark.skipif(sys.platform == 'linux', reason="does not run on linux")
def test_Pipeline():
    # test on a single stream
    stream_config = {
        "generator": generator.normal_dist,
        'kwargs': {
            "grange": 8,
            "buffer_size": 100,
            "sleep_interval": 0,
            "sigma": 1,
            "sr": 80
        }
    }

    window_size = 12.8
    start_time = datetime.now()
    stream = GeneratorSlidingWindowStream(
        stream_config, window_size=window_size, name='stream-1')

    pipeline = Pipeline(max_processes=2, scheduler='threads',
                        name='single-stream-pipeline')
    pipeline.add_stream(stream)
    pipeline.set_processor(_pipeline_test_processor)
    pipeline.start(start_time=start_time, process_start_time=start_time)

    results = []
    for result, st, et, prev_st, prev_et, name in pipeline.get_iterator():
        result['WINDOW_ST'] = st
        result['WINDOW_ET'] = et
        result['PREV_WINDOW_ST'] = prev_st
        result['PREV_WINDOW_ET'] = prev_et
        result['STREAM_NAME'] = name
        results.append(result)
        if len(results) == 5:
            break
    pipeline.stop()
    results = pd.concat(results, axis=0, sort=False)
    durations = (results['WINDOW_ET'] -
                 results['WINDOW_ST']) / pd.Timedelta(1, unit='S')
    np.testing.assert_array_almost_equal(durations, window_size, decimal=1)

    # test on three streams, using the same pipeline but change the parameter of the first stream a little bit
    stream2_config = {
        "generator": generator.normal_dist,
        'kwargs': {
            "grange": 4,
            "buffer_size": 400,
            "sleep_interval": 1,
            "sigma": 2,
            "sr": 50
        }
    }

    start_time = datetime.now()
    stream2 = GeneratorSlidingWindowStream(
        stream2_config, window_size=window_size, name='stream-2')
    pipeline.add_stream(stream2)
    pipeline.start(start_time=start_time, process_start_time=start_time)

    results = []
    for result, st, et, prev_st, prev_et, name in pipeline.get_iterator():
        result['WINDOW_ST'] = st
        result['WINDOW_ET'] = et
        result['PREV_WINDOW_ST'] = prev_st
        result['PREV_WINDOW_ET'] = prev_et
        result['STREAM_NAME'] = name
        np.testing.assert_array_equal(
            result['START_TIME'].values, result['START_TIME'].values[0])
        results.append(result)
        if len(results) == 5:
            break
    pipeline.stop()
    results = pd.concat(results, axis=0, sort=False)
    durations = (results['WINDOW_ET'] -
                 results['WINDOW_ST']) / pd.Timedelta(1, unit='S')
    np.testing.assert_array_almost_equal(durations, window_size, decimal=1)



"""arus command line application

Usage:
  arus signaligner FOLDER PID [SR] [-t <file_type>] [--date_range=<date_range>] [--auto_range=<auto_range>] [--debug]
  arus app APP_COMMAND FOLDER NAME [--app_version=<app_version>]
  arus dataset DATASET_COMMAND DATASET_NAME [FOLDER] [OUTPUT_FOLDER] [--debug]
  arus package PACK_COMMAND [NEW_VERSION] [--dev] [--release] [--debug]
  arus --help
  arus --version

Arguments:
  FOLDER                                        Dataset folder.
  PID                                           Participant ID.
  SR                                            Sampling rate in Hz.
  APP_COMMAND                                   Sub commands for app command. Either "build" or "run".
  NAME                                          Name of the app.
  PACK_COMMAND                                  "release", "docs"
  NEW_VERSION                                   "major", "minor", "patch" or number.

Options:
  -t <file_type>, --file_type=<file_type>       File type: either "sensor" or "annotation". If omit, both are included.
  --date_range=<date_range>                     Date range. E.g., "--date_range 2020-06-01,2020-06-10", or "--date_range 2020-06-01," or "--date_range ,2020-06-10".
  --auto_range=<auto_range>                     Auto date freq. Default is "W-SUN", or weekly starting from Sunday.
  --app_version=<app_version>                   App version. If omit, default is the same as the version of arus package.
  -h, --help                                    Show help message.
  -v, --version                                 Program/app version.
"""

from docopt import docopt
from loguru import logger
from . import developer as dev
from . import dataset as ds
from . import mhealth_format as mh
from . import plugins
from . import models
from . import spades_lab as slab
from . import env
from .plugins import signaligner
import glob
from datetime import datetime, timedelta
import os
import tqdm
import sys
import subprocess
import json
import pprint
import pandas as pd
import math
import shutil
import pkg_resources


def cli():
    arguments = docopt(__doc__, version='arus {}'.format(
        dev._find_current_version(".", "arus")))
    if arguments['--debug']:
        logger.remove()
        logger.add(sys.stderr, level='DEBUG')
    else:
        logger.remove()
        logger.add(sys.stderr, level='INFO')
    logger.debug(arguments)

    if arguments['signaligner']:
        signaligner_command(arguments)
    elif arguments['app']:
        app_command(arguments)
    elif arguments['dataset']:
        dataset_command(arguments)
    elif arguments['package']:
        package_command(arguments)


def signaligner_command(arguments):
    pid = arguments['PID']
    root = arguments['FOLDER']
    file_type = arguments['--file_type']
    date_range = arguments['--date_range'].split(
        ',') if arguments['--date_range'] is not None else None
    # auto range would be weekly files starting from Sunday
    auto_range = arguments['--auto_range'] or 'W-SUN'
    sr = int(arguments['SR']) if arguments['SR'] is not None else None

    logger.debug('Signaligner arguments: {}', [
                 root, pid, file_type, sr, date_range, auto_range])

    if file_type == 'sensor':
        convert_to_signaligner(root, pid, file_type, sr=sr,
                               date_range=date_range, auto_range=auto_range)
    elif file_type == 'annotation':
        convert_to_signaligner(root, pid, file_type, sr=sr,
                               date_range=date_range, auto_range=auto_range)
    elif file_type is None:
        convert_to_signaligner_both(
            root, pid, sr=sr, date_range=date_range, auto_range=auto_range)


def convert_to_signaligner_both(root, pid, sr, date_range=None, auto_range='W-SUN'):
    convert_to_signaligner(root, pid, 'annotation', sr, date_range, auto_range)
    convert_to_signaligner(root, pid, 'sensor', sr, date_range, auto_range)


def convert_to_signaligner(root, pid, file_type, sr=None, date_range=None, auto_range='W-SUN'):
    session_span = mh.get_session_span(pid, root)
    session_span = signaligner.shrink_session_span(session_span, date_range)
    logger.info('Session span: {}'.format(session_span))

    if file_type == 'annotation':
        filepaths = mh.get_annotation_files(pid, root)
        def filter_fun(f): return group == mh.parse_data_id_from_filepath(f)
    elif file_type == 'sensor':
        assert sr is not None
        filepaths = mh.get_sensor_files(pid, root)
        logger.debug(session_span)
        def filter_fun(f): return group in f
    groups = set(map(mh.parse_data_id_from_filepath, filepaths))
    logger.debug(groups)
    for group in groups:
        logger.info('Convert files of {}'.format(group))
        group_files = list(filter(filter_fun, filepaths))
        logger.debug(group_files)
        if file_type == 'sensor':
            sensor_type = mh.parse_sensor_type_from_filepath(group_files[0])
        sub_session_markers = signaligner.auto_split_session_span(
            session_span, auto_range)
        logger.debug(sub_session_markers)
        for i in range(len(sub_session_markers) - 1):
            sub_session_span = sub_session_markers[i:i + 2]
            st_display = sub_session_span[0].strftime('%Y%m%d%H')
            et_display = sub_session_span[1].strftime('%Y%m%d%H')
            logger.info(
                f'Process sub session: {st_display} - {et_display} based on {auto_range}')

            # call signify functions
            if file_type == 'annotation':
                output_path = os.path.join(
                    root, pid, mh.DERIVED_FOLDER, 'signaligner', f'{pid}_{st_display}_{et_display}_labelsets', f'{pid}_{group}_{st_display}_{et_display}.{file_type}.csv')
                os.makedirs(os.path.dirname(output_path), exist_ok=True)
                signaligner.signify_annotation_files(
                    group_files, group, output_path, sub_session_span)
            elif file_type == 'sensor':
                # set output file paths
                output_path = os.path.join(
                    root, pid, mh.DERIVED_FOLDER, 'signaligner', f'{pid}_{st_display}_{et_display}_sensors', f'{pid}_{sensor_type}_{group}_{st_display}_{et_display}.{file_type}.csv')
                os.makedirs(os.path.dirname(output_path), exist_ok=True)
                output_annotation_path = os.path.join(os.path.dirname(
                    output_path.replace('sensors', 'labelsets')), f'{pid}_missing_{group}_{st_display}_{et_display}.annotation.csv')
                os.makedirs(os.path.dirname(
                    output_annotation_path), exist_ok=True)
                signaligner.signify_sensor_files(
                    group_files, group, output_path, output_annotation_path, sub_session_span, sr)


def app_command(arguments):
    if arguments['APP_COMMAND'] == 'build':
        root = arguments['FOLDER']
        name = arguments['NAME']
        version = arguments['--app_version'] or dev._find_current_version(
            '.', 'arus')
        dev.build_arus_app(root, name, version)
    elif arguments['APP_COMMAND'] == 'run':
        root = arguments['FOLDER']
        name = arguments['NAME']
        subprocess.run(['python', os.path.join(
            root, 'apps', name, 'main.py')], shell=True)


def dataset_command(arguments):
    if arguments['DATASET_COMMAND'] == 'compress':
        root = arguments['FOLDER']
        name = arguments['DATASET_NAME']
        output_folder = arguments['OUTPUT_FOLDER']
        logger.info(
            f'Compressing dataset as {name}.tar.gz in folder {output_folder}')
        dev.compress_dataset(root, output_folder, f'{name}.tar.gz')
    elif arguments['DATASET_COMMAND'] == 'query':
        name = arguments['DATASET_NAME']
        if name == 'all':
            logger.info("Available sample data names")
            logger.info(ds.get_available_sample_data())
        else:
            logger.info(f'Query dataset: {name} in ARUS package')
            logger.info(ds.get_sample_datapath(name))


def format_time(ts):
    s = ts.strftime('%Y-%m-%d %H:%M:%S')
    ms = '{:03.0f}'.format(math.floor(ts.microsecond / 1000.0))
    result = "{}.{}".format(s, ms)
    return result


def package_command(arguments):
    if arguments['PACK_COMMAND'] == 'release':
        is_dev = arguments['--dev'] or False
        nver = arguments['NEW_VERSION']
        assert nver is not None
        release = arguments['--release'] or False
        _release_package(nver, is_dev=is_dev, release=release)
    elif arguments['PACK_COMMAND'] == 'docs':
        is_dev = arguments['--dev'] or True
        release = arguments['--release'] or False
        if is_dev and not release:
            dev.dev_website()
        elif release:
            dev.build_website()


def _release_package(nver, is_dev=False, release=False):
    new_version = dev.bump_package_version('.', 'arus', nver, is_dev)
    if new_version is not None and dev.command_is_available('git') and release:
        dev.commit_repo(f"Bump version to {new_version}")
        dev.tag_repo(new_version)
        dev.push_repo()
        dev.push_tag(new_version)



"""
Module of testing utility functions

This module provides utility functions for testing modules in `arus.core`.

Author: Qu Tang

Date: 2019-11-15

License: see LICENSE file
"""

import os
import glob


def load_test_data(file_type='mhealth', sensor_type='sensor', file_num='multiple', exception_type='consistent_sr'):
    """Load test data that are bundled in the package.

    Args:
        file_type (str, optional): either files stored in 'mhealth' or 'actigraph' csv format. Defaults to 'mhealth'.
        sensor_type (str, optional): {'sensor', 'annotation', 'feature', 'class_labels'}. Defaults to 'sensor'.
        file_num (str, optional): either 'single' or 'multiple' files. Defaults to 'multiple'.
        exception_type (str, optional): 'consistent_sr' (consistent sampling rate), 'inconsistent_sr' (inconsistent sampling rate), 'missing' (with missing data), 'no_missing' (with no missing data), 'multi_placements' (features from multiple placements), 'single_placement' (features from single placement), 'multi_tasks' (multiple class columns), 'single_task' (single class column). Defaults to 'consistent_sr'.

    Returns:
        tuple: the first element is the list of filepaths found and sorted, the second element is the sampling rate if `sensor_type` is 'sensor', otherwise it is `None`.
    """

    cwd = os.path.dirname(os.path.realpath(__file__))
    data_folder = os.path.join(cwd, '..', 'data', file_num,
                               file_type, sensor_type, exception_type)
    meta_file = os.path.join(data_folder, 'meta.csv')
    if file_num == 'multiple':
        data = list(filter(lambda f: 'meta.csv' != os.path.basename(f), glob.glob(
            os.path.join(data_folder, '*.csv*'))))
        data = sorted(data)
    else:
        data = list(filter(lambda f: 'meta.csv' != os.path.basename(f), glob.glob(
            os.path.join(data_folder, '*.csv*'))))[0]
    if sensor_type == 'sensor':
        with open(meta_file, 'r') as meta:
            sr = int(meta.readline().split(',')[1])
    else:
        sr = None
    return data, sr



import queue
import PySimpleGUI as sg
import enum
from loguru import logger


class BaseEvent(enum.Enum):
    CLOSE_WINDOW = enum.auto()


class BaseWindow:
    def __init__(self, title, column_width):
        self._title = title
        self._column_width = column_width
        self._events = queue.Queue()

    def start(self):
        self._state = self.init_states()
        self._window = self.init_views()
        while True:
            self.update_states_and_events()
            if self.dispatch_events():
                break
        self._window.close()

    def init_states(self):
        raise NotImplementedError(
            "Sub class must implements this method to initialize the app state before starting the window. This method should return an app.App_State object.")

    def init_views(self):
        raise NotImplementedError(
            "Sub class must implements this method to initialize the views of the window before starting the window. This method should return a sg.Window object.")

    def update_states_and_events(self):
        inherited_window_event, new_values = self._window.read(timeout=20)
        if inherited_window_event == sg.TIMEOUT_KEY:
            pass
        elif inherited_window_event is not None:
            self._events.put(inherited_window_event)

        if new_values is not None:
            self._update_states_and_events(inherited_window_event, new_values)

        if inherited_window_event is None:
            self._events.put(BaseEvent.CLOSE_WINDOW)

    def dispatch_events(self):
        for event in self._events.queue:
            logger.info('Dispatch event: ' + str(event))
            if event == BaseEvent.CLOSE_WINDOW:
                return True
            else:
                self._dispatch_events(event)
        self._events.queue.clear()
        return False

    def _update_states_and_events(self, event, values):
        raise NotImplementedError(
            'Sub class must implements this method to update states and events based on current app state and input events.')

    def _dispatch_events(self, event):
        raise NotImplementedError(
            "Sub class must implements this method to handle different events.")



import enum
from loguru import logger
import queue
import time
import datetime as dt
import copy

import PySimpleGUI as sg
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import numpy as np
import pandas as pd

import app_state as app
import backend
import base
import components as comp
import arus


class Event(enum.Enum):
    CONNECT_DEVICES = enum.auto()
    DEVICE_CONNECTED = enum.auto()
    DISCONNECT_DEVICES = enum.auto()
    DEVICE_DISCONNECTED = enum.auto()
    MODE_CHANGED = enum.auto()
    SCAN_DEVICES = enum.auto()
    SCAN_COMPLETED = enum.auto()
    START_TEST = enum.auto()
    TEST_STARTED = enum.auto()
    STOP_TEST = enum.auto()
    TEST_STOPPED = enum.auto()
    INFERENCE_UPDATE = enum.auto()
    TEST_LABEL_CHANGED = enum.auto()
    TIMER_UPDATE = enum.auto()
    GUIDANCE_PLAYED = enum.auto()
    SAVE_ANNOTATION = enum.auto()
    ANNOTATION_SAVED = enum.auto()


class StreamWindow(base.BaseWindow):
    def __init__(
        self,
        title,
        model=None,
        labels=None,
        support_modes=[
            backend.PROCESSOR_MODE.TEST_ONLY,
            backend.PROCESSOR_MODE.TEST_AND_SAVE
        ]
    ):
        super().__init__(title, 100)
        self._test_model = model
        self._labels = labels
        self._support_modes = support_modes
        self._selected_mode = None
        self._selected_placements = None
        self._selected_devices = ['', '', '']
        self._selected_placement_addrs = None

        self._scan_task = None
        self._connect_task = None
        self._disconnect_task = None
        self._start_task = None
        self._stop_task = None
        self._guide_task = None
        self._save_task = None

        self._pipeline = None
        self._latest_inference = None
        self._inference_queue = []

        self._result = None

        self._selected_test_label = None
        self._timer = comp.Timer()

        self._current_annotation = {
            "HEADER_TIME_STAMP": None, "START_TIME": None, "STOP_TIME": None, "LABEL_NAME": None
        }
        self._annotation_save_queue = queue.Queue()

        self._current_guidance = {
            "HEADER_TIME_STAMP": None, "START_TIME": None, "STOP_TIME": None, "LABEL_NAME": None
        }

    def get_result(self):
        return self._result

    def init_states(self):
        state = app.AppState.getInstance()
        if self._test_model is not None:
            self._labels = self._labels or self._test_model[0].classes_.tolist(
            )
            self._selected_mode = self._support_modes[0]
            self._selected_placements = self._test_model[-2][:]
        else:
            self._selected_mode = backend.PROCESSOR_MODE.COLLECT_ONLY
            self._selected_placements = []
        if self._selected_mode in [
                backend.PROCESSOR_MODE.COLLECT_ONLY,
                backend.PROCESSOR_MODE.TEST_AND_COLLECT,
                backend.PROCESSOR_MODE.ACTIVE_COLLECT]:
            self._result = state.new_dataset
        else:
            self._result = None
        return state

    def init_views(self):
        header_row = [
            comp.heading(
                self._title,
                fixed_column_width=self._column_width
            )
        ]

        self._scan_button = comp.control_button(
            "Scan devices",
            disabled=False,
            key=Event.SCAN_DEVICES
        )
        scan_button_row = [
            self._scan_button
        ]

        device_info_row = self.init_devices_view()

        self._connect_button = comp.control_button(
            "Connect devices",
            disabled=True,
            key=Event.CONNECT_DEVICES
        )

        self._disconnect_button = comp.control_button(
            "Disconnect devices",
            disabled=True,
            key=Event.DISCONNECT_DEVICES
        )

        self._start_button = comp.control_button(
            "Start " + self._title,
            disabled=True,
            key=Event.START_TEST
        )

        self._stop_button = comp.control_button(
            "Stop " + self._title,
            disabled=True,
            key=Event.STOP_TEST
        )

        control_mode_row = self.init_mode_controls()

        self._timer_view = comp.text('00:00:00')
        self._guidance_view = comp.text('', fixed_column_width=25)
        info_row = [
            self._timer_view, self._guidance_view
        ]

        button_row = [
            self._connect_button, self._disconnect_button, self._start_button, self._stop_button
        ]

        test_info_row = self.init_test_view()

        data_summary_row = self.init_data_summary_view()

        layout = [
            header_row,
            scan_button_row,
            device_info_row,
            button_row,
            control_mode_row,
            info_row,
            test_info_row,
            data_summary_row
        ]

        return sg.Window(
            self._title,
            layout=layout,
            finalize=True
        )

    def _update_states_and_events(self, event, values):
        if self._scan_task is not None:
            if self._scan_task.ready():
                self._state.nearby_devices = sorted(self._scan_task.get())
                self._selected_devices = self._state.nearby_devices[:3]
                self._events.put(Event.SCAN_COMPLETED)
                self._scan_task = None

        if self._connect_task is not None:
            if self._connect_task.ready():
                self._pipeline = self._connect_task.get()
                self._events.put(Event.DEVICE_CONNECTED)
                self._connect_task = None

        if self._disconnect_task is not None:
            if self._disconnect_task.ready():
                self._events.put(Event.DEVICE_DISCONNECTED)
                self._disconnect_task = None

        if self._start_task is not None:
            if self._start_task.ready():
                self._events.put(Event.TEST_STARTED)
                self._start_task = None

        if self._stop_task is not None:
            if self._stop_task.ready():
                self._events.put(Event.TEST_STOPPED)
                self._stop_task = None

        if self._guide_task is not None:
            if self._guide_task.ready():
                self._events.put(Event.GUIDANCE_PLAYED)
                self._guide_task = None

        if self._save_task is not None:
            if self._save_task.done():
                self._events.put(Event.ANNOTATION_SAVED)
                self._save_task = None

        for mode in self._support_modes:
            if values[mode]:
                if mode != self._selected_mode:
                    if mode != backend.PROCESSOR_MODE.COLLECT_ONLY:
                        self._selected_placements = self._test_model[-2][:]

                    self._events.put(Event.MODE_CHANGED)
                    self._selected_mode = mode
                    break

        if event == 'DW_ADDR':
            self._selected_devices[0] = values[event]
        if event == 'DA_ADDR':
            self._selected_devices[1] = values[event]
        if event == 'DT_ADDR':
            self._selected_devices[2] = values[event]

        if event in ['DW', 'DA', 'DT']:
            if values[event]:
                self._selected_placements.append(event)
            else:
                self._selected_placements.remove(event)
            logger.info('Selected placements: ' +
                        str(self._selected_placements))

        if self._state.nearby_devices is not None and len(self._selected_devices) >= 3:
            self._selected_placement_addrs = [
                self._selected_devices[
                    ['DW', 'DA', 'DT'].index(placement)
                ] for placement in self._selected_placements
            ]

        if self._pipeline is not None:
            self._latest_inference, st, et, _, _, _ = next(
                self._pipeline.get_iterator(timeout=0.02))
            if self._latest_inference is not None:
                logger.info(self._latest_inference)
                self.format_inference_result()
                self.add_to_result()
                self._events.put(Event.INFERENCE_UPDATE)

        if event == Event.START_TEST:
            self._timer.start()
            if self._selected_mode != backend.PROCESSOR_MODE.TEST_ONLY:
                self._annotation_save_queue.queue.clear()
                self.start_new_annotation()

        if event == Event.STOP_TEST:
            self._timer.stop()
            if self._selected_mode != backend.PROCESSOR_MODE.TEST_ONLY:
                self.complete_new_annotation()
                self._events.put(Event.SAVE_ANNOTATION)

        if event == Event.DISCONNECT_DEVICES:
            self._timer.stop()
            self._timer.reset()

        if len(values[Event.TEST_LABEL_CHANGED]) > 0:
            if self._selected_test_label != values[Event.TEST_LABEL_CHANGED][0] and self._selected_mode != backend.PROCESSOR_MODE.TEST_ONLY and self._timer.is_running():
                self.complete_new_annotation()
                self._events.put(Event.SAVE_ANNOTATION)
                self._selected_test_label = values[Event.TEST_LABEL_CHANGED][0]
                self.start_new_annotation()
            elif self._selected_test_label != values[Event.TEST_LABEL_CHANGED][0]:
                self._selected_test_label = values[Event.TEST_LABEL_CHANGED][0]

        if self._timer.is_running() and self._timer.since_last_tick() >= 1:
            self._events.put(Event.TIMER_UPDATE)
            self._timer.tick()

    def _dispatch_events(self, event):
        if event == Event.SCAN_DEVICES:
            self.scan_devices()
            self.disable_scan_button()
        elif event == Event.SCAN_COMPLETED:
            self.update_device_addrs()
            self.enable_scan_button()
            self.enable_connect_button()
        elif event == Event.CONNECT_DEVICES:
            if len(set(self._selected_devices)) != len(self._selected_devices):
                sg.Popup(
                    'Please make sure the selected device addresses are all different', title='Warning')
            else:
                self.connect_devices()
                self.disable_connect_button()
                self.disable_device_selections()
        elif event == Event.DEVICE_CONNECTED:
            self.enable_disconnect_button()
            self.disable_connect_button(finished=True)
            self.enable_start_button()
            self._control_modes.disable_all()
        elif event == Event.DISCONNECT_DEVICES:
            self.disconnect_devices()
            self.disable_disconnect_button()
            self.disable_start_button()
        elif event == Event.DEVICE_DISCONNECTED:
            self.enable_connect_button()
            self.enable_device_selections()
            self.disable_disconnect_button(finished=True)
            self.disable_start_button(finished=True)
            self.disable_stop_button(finished=True)
            self._control_modes.enable_all()
        elif event == Event.START_TEST:
            self.start_test()
            self.disable_start_button()
            self.disable_disconnect_button()
        elif event == Event.TEST_STARTED:
            self.disable_start_button(finished=True)
            self.enable_stop_button()
        elif event == Event.STOP_TEST:
            self.stop_test()
            self.disable_stop_button()
        elif event == Event.TEST_STOPPED:
            self.disable_stop_button(finished=True)
            self.enable_start_button()
            self.enable_disconnect_button()
        elif event == Event.INFERENCE_UPDATE:
            if self._selected_mode != backend.PROCESSOR_MODE.COLLECT_ONLY:
                self.update_test_view()
            if self._selected_mode == backend.PROCESSOR_MODE.ACTIVE_COLLECT:
                self.play_active_guidance()
            self.update_data_summary_view()
        elif event == Event.TIMER_UPDATE:
            self.update_timer_view()
        elif event == Event.GUIDANCE_PLAYED:
            self._guidance_view.update('')
        elif event == Event.SAVE_ANNOTATION:
            self.save_annotation()
        elif event == Event.ANNOTATION_SAVED:
            logger.info('Annotation is saved.')
        elif event == Event.MODE_CHANGED:
            if self._selected_mode == backend.PROCESSOR_MODE.COLLECT_ONLY:
                self.enable_placement_selections()
            else:
                self.update_placement_selections()
                self.disable_placement_selections()

    def init_devices_view(self):
        if 'DW' in self._selected_placements and self._selected_mode != backend.PROCESSOR_MODE.COLLECT_ONLY:
            dw_selected = True
            dw_disabled = True
        elif self._selected_mode == backend.PROCESSOR_MODE.COLLECT_ONLY:
            dw_selected = False
            dw_disabled = False
        else:
            dw_selected = False
            dw_disabled = True
        self._info_wrist = comp.DeviceInfo(device_name='Dominant Wrist',
                                           placement_img_url='./assets/dom_wrist.png', fixed_column_width=30, device_name_key="DW",
                                           device_addr_key="DW_ADDR",
                                           device_selected=dw_selected,
                                           device_selection_disabled=dw_disabled)
        if 'DA' in self._selected_placements and self._selected_mode != backend.PROCESSOR_MODE.COLLECT_ONLY:
            da_selected = True
            da_disabled = True
        elif self._selected_mode == backend.PROCESSOR_MODE.COLLECT_ONLY:
            da_selected = False
            da_disabled = False
        else:
            da_selected = False
            da_disabled = True
        self._info_ankle = comp.DeviceInfo(device_name='Right Ankle',
                                           placement_img_url='./assets/right_ankle.png',
                                           fixed_column_width=30, device_name_key="DA",
                                           device_addr_key="DA_ADDR",
                                           device_selected=da_selected, device_selection_disabled=da_disabled)
        if 'DT' in self._selected_placements and self._selected_mode != backend.PROCESSOR_MODE.COLLECT_ONLY:
            dt_selected = True
            dt_disabled = True
        elif self._selected_mode == backend.PROCESSOR_MODE.COLLECT_ONLY:
            dt_selected = False
            dt_disabled = False
        else:
            dt_selected = False
            dt_disabled = True
        self._info_thigh = comp.DeviceInfo(device_name='Dominant Thigh',
                                           placement_img_url='./assets/dom_thigh.png',
                                           fixed_column_width=30, device_name_key="DT",
                                           device_addr_key='DT_ADDR',
                                           device_selected=dt_selected, device_selection_disabled=dt_disabled)
        return [sg.Column(layout=self._info_wrist.get_component(), scrollable=False), sg.Column(layout=self._info_ankle.get_component(), scrollable=False), sg.Column(layout=self._info_thigh.get_component(), scrollable=False)]

    def init_test_view(self):
        if self._selected_mode in [
            backend.PROCESSOR_MODE.TEST_AND_SAVE,
            backend.PROCESSOR_MODE.TEST_ONLY
        ]:
            extra_label = 'Any'
        elif self._selected_mode in [
            backend.PROCESSOR_MODE.ACTIVE_COLLECT,
            backend.PROCESSOR_MODE.COLLECT_ONLY,
            backend.PROCESSOR_MODE.TEST_AND_COLLECT
        ]:
            extra_label = 'Sync'
        if self._test_model is not None:
            model_labels = self._test_model[0].classes_.tolist()
        else:
            model_labels = self._labels
        n = len(model_labels)

        self._test_label_list = comp.selection_list(
            [extra_label] + self._labels,
            default_selections=extra_label,
            mode='single',
            fixed_column_width=int(self._column_width * 1 / 6),
            rows=8,
            key=Event.TEST_LABEL_CHANGED
        )

        self._labelgrid = comp.LabelGrid(
            n,
            fixed_column_width=int(self._column_width * 5 / 6),
            default_background_color='green',
            default_text_color='white',
            n_cols=4
        )
        self._labelgrid.add_labels(
            model_labels, keys=[None] * n, background_colors=[None] * n, text_colors=[None] * n)
        return [
            self._test_label_list,
            sg.Column(
                layout=self._labelgrid.get_component(),
                scrollable=False
            )
        ]

    def init_data_summary_view(self):
        cells = backend.get_data_summary_table(
            self._result)
        cells = cells or [["", "", ""]]
        self._data_summary = comp.table(
            cells, headings=["Activity", "# of windows", "Accuracy"], fixed_column_width=self._column_width)
        return [self._data_summary]

    def init_mode_controls(self):
        mode_names = [mode.name for mode in self._support_modes]
        disable_states = [False] * len(self._support_modes)
        default_index = 0
        if self._test_model is None:
            i = self._support_modes.index(
                backend.PROCESSOR_MODE.ACTIVE_COLLECT
            )
            if default_index == i:
                default_index += 1
            disable_states[i] = True
            i = self._support_modes.index(
                backend.PROCESSOR_MODE.TEST_AND_COLLECT
            )
            if default_index == i:
                default_index += 1
            disable_states[i] = True

        self._control_modes = comp.RadioGroup(
            'PROCESS_MODE',
            len(self._support_modes),
            default_index=default_index,
            fixed_column_width=self._column_width
        )
        self._control_modes.add_radioboxes(
            mode_names,
            keys=self._support_modes,
            disable_states=disable_states
        )
        return self._control_modes.get_component()

    def scan_devices(self):
        self._scan_task = backend.get_nearby_devices()

    def connect_devices(self):
        self._connect_task = backend.connect_devices(
            self._selected_placement_addrs, model=self._test_model, mode=self._selected_mode, output_folder=self._state.output_folder, pid=self._state.pid, placement_names=self._selected_placements)

    def start_test(self):
        self._start_task = backend.start_test_model(self._pipeline)

    def stop_test(self):
        self._stop_task = backend.stop_test_model(self._pipeline)

    def play_active_guidance(self):
        if (np.all(self._inference_queue[-2:]) and len(self._inference_queue) > 1) or len(self._inference_queue) >= 4:
            text = 'Switch now!'
            self._guide_task = backend.play_sound('./assets/switch_now')
            self._inference_queue.clear()
        else:
            text = 'Keep going!'
            self._guide_task = backend.play_sound('./assets/keep_going')
        self._guidance_view.update(text)

    def disconnect_devices(self):
        self._disconnect_task = backend.disconnect_devices(self._pipeline)

    def save_annotation(self):
        annotation = self._annotation_save_queue.get()
        session_name = None
        if self._selected_mode == backend.PROCESSOR_MODE.COLLECT_ONLY:
            session_name = 'PassiveSensing'
        elif self._selected_mode == backend.PROCESSOR_MODE.TEST_AND_COLLECT:
            session_name = 'PassiveSensing'
        elif self._selected_mode == backend.PROCESSOR_MODE.ACTIVE_COLLECT:
            session_name = 'ActiveTraining'
        elif self._selected_mode == backend.PROCESSOR_MODE.TEST_AND_SAVE:
            if self._test_model == self._state.origin_model:
                session_name = 'TestOriginModel'
            elif self._test_model == self._state.new_model:
                session_name = 'TestNewModel'
        self._save_task = backend.save_annotation(
            annotation,
            output_folder=self._state.output_folder,
            pid=self._state.pid,
            session_name=session_name
        )
        logger.info('Saving annotation...')

    def disable_scan_button(self):
        self._scan_button.update('Scanning...', disabled=True)

    def enable_scan_button(self):
        self._scan_button.update('Scan devices', disabled=False)

    def update_device_addrs(self):
        self._info_wrist.update_addr_list(self._state.nearby_devices, index=0)
        self._info_ankle.update_addr_list(self._state.nearby_devices, index=1)
        self._info_thigh.update_addr_list(self._state.nearby_devices, index=2)

    def add_to_result(self):
        if self._result is None:
            self._result = self._latest_inference[1]
        else:
            self._result = pd.concat([self._result,
                                      self._latest_inference[1]], sort=False
                                     )
        self._result.reset_index(drop=True)

    def parse_inference_result(self):
        correct = self._latest_inference[0][0].split(
            ':')[0] == self._selected_test_label or self._selected_test_label == 'Any'
        return correct

    def update_test_view(self):
        correct = self.parse_inference_result()
        self._inference_queue.append(correct)
        base_color = 'g' if correct else 'r'
        new_b_colors = []
        new_t_colors = []
        for value in self._latest_inference[0]:
            name = value.split(':')[0]
            num = float(value.split(':')[1])
            if not correct and name == self._selected_test_label:
                new_color = 'blue'
                text_color = 'white'
            else:
                new_color = arus.ext.plotting.adjust_lightness(
                    base_color, amount=num / 100.0, return_format='hex')
                text_color = "white" if num > 20 else "gray"
            new_b_colors.append(new_color)
            new_t_colors.append(text_color)
        self._labelgrid.update_labels(
            self._latest_inference[0], new_b_colors, new_t_colors)

    def update_data_summary_view(self):
        cells = backend.get_data_summary_table(
            self._result)
        cells = cells or [["", "", ""]]
        self._data_summary.update(cells)

    def update_timer_view(self):
        self._timer_view.update(
            self._timer.get_total_lapsed_time(formatted=True) + ' seconds')

    def format_inference_result(self):
        self._latest_inference = list(self._latest_inference)
        # sort probs
        if self._selected_mode != backend.PROCESSOR_MODE.COLLECT_ONLY:
            model_labels = self._test_model[0].classes_.tolist()
            pred_probs = self._latest_inference[0][0].tolist()
            formatted = []
            for label, value in zip(model_labels, pred_probs):
                formatted.append(
                    label + ': ' + str(int(round(value, 2) * 100)))
            sort_formatted = list(sorted(
                formatted, key=lambda v: int(v.split(':')[1]), reverse=True))
            self._latest_inference[0] = sort_formatted
            # add labels to feature dataframe
            self._latest_inference[1]['GT_LABEL'] = self._selected_test_label
            self._latest_inference[1]['PREDICTION'] = self._latest_inference[0][0].split(":")[
                0]
            self._latest_inference[1]['PID'] = self._state.pid
        else:
            self._latest_inference[1]['GT_LABEL'] = self._selected_test_label
            self._latest_inference[1]['PID'] = self._state.pid

    def disable_connect_button(self, finished=False):
        if finished:
            text = 'Connected'
        else:
            text = 'Connecting...'
        self._connect_button.update(text, disabled=True)

    def disable_disconnect_button(self, finished=False):
        if finished:
            text = 'Disconnected'
        else:
            text = 'Disconnecting...'
        self._disconnect_button.update(text, disabled=True)

    def enable_connect_button(self):
        self._connect_button.update('Connect devices', disabled=False)

    def enable_disconnect_button(self):
        self._disconnect_button.update('Disconnect', disabled=False)

    def enable_start_button(self):
        self._start_button.update('Start test', disabled=False)

    def disable_start_button(self, finished=False):
        if finished:
            text = 'Started'
        else:
            text = 'Starting...'
        self._start_button.update(text, disabled=True)

    def enable_stop_button(self):
        self._stop_button.update('Stop test', disabled=False)

    def disable_stop_button(self, finished=False):
        if finished:
            text = 'Stopped'
        else:
            text = 'Stopping...'
        self._stop_button.update(text, disabled=True)

    def start_new_annotation(self):
        start_time = dt.datetime.now()
        self._current_annotation['LABEL_NAME'] = [self._selected_test_label]
        self._current_annotation['START_TIME'] = [start_time]
        self._current_annotation['HEADER_TIME_STAMP'] = [start_time]
        self._current_annotation['STOP_TIME'] = None

    def complete_new_annotation(self):
        stop_time = dt.datetime.now()
        self._current_annotation['STOP_TIME'] = [stop_time]
        self._annotation_save_queue.put(
            copy.deepcopy(self._current_annotation))

    def is_annotation_incomplete(self):
        return self._current_annotation['STOP_TIME'] is None

    def enable_placement_selections(self):
        self._info_wrist.enable_selection()
        self._info_ankle.enable_selection()
        self._info_thigh.enable_selection()

    def disable_placement_selections(self):
        self._info_wrist.disable_selection()
        self._info_ankle.disable_selection()
        self._info_thigh.disable_selection()

    def update_placement_selections(self):
        selected = 'DW' in self._selected_placements
        self._info_wrist.update_selection(selected)
        selected = 'DA' in self._selected_placements
        self._info_ankle.update_selection(selected)
        selected = 'DT' in self._selected_placements
        self._info_thigh.update_selection(selected)

    def disable_device_selections(self):
        self._info_wrist.disable_addr_list()
        self._info_ankle.disable_addr_list()
        self._info_thigh.disable_addr_list()

    def enable_device_selections(self):
        self._info_wrist.enable_addr_list()
        self._info_ankle.enable_addr_list()
        self._info_thigh.enable_addr_list()



import matplotlib.colors as mc
import colorsys
from matplotlib import rcParams, rcParamsDefault
from matplotlib import pyplot as plt
import seaborn as sns
import numpy as np


def adjust_lightness(color, amount=0.5, return_format='rgb'):
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    result = colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])
    if return_format == 'rgb':
        return result
    elif return_format == 'hex':
        return mc.to_hex(result)
    else:
        return result


def set_for_publication():
    rcParams.update(rcParamsDefault)
    rcParams['font.family'] = 'serif'
    rcParams['font.size'] = 10
    rcParams['font.serif'] = ['Times New Roman']


def confusion_matrix(cm_df):
    rcParams.update(rcParamsDefault)
    rcParams['font.family'] = 'serif'
    rcParams['font.size'] = 10
    rcParams['font.serif'] = ['Times New Roman']
    fig, ax = plt.subplots(figsize=(8, 8))
    sns.set_style({
        'font.family': 'serif',
        'font.size': 10
    })
    g = sns.heatmap(cm_df, annot=True, cmap="Greys",
                    cbar=False, fmt='d', robust=True, linewidths=0.2)
    g.set(xlabel="Prediction", ylabel="Ground truth")
    plt.tight_layout()
    return fig


def learning_curve(lc_df, title='Learning curve', axes=None, ylim=None):
    if axes is None:
        fig, axes = plt.subplots(1, 3, figsize=(20, 5))

    axes[0].set_title(title)
    if ylim is not None:
        axes[0].set_ylim(*ylim)
    axes[0].set_xlabel("Training examples")
    axes[0].set_ylabel("Score")

    train_sizes = lc_df['train_sizes']
    train_scores_mean = lc_df['train_scores_means']
    train_scores_std = lc_df['train_scores_stds']
    test_scores_mean = lc_df['test_scores_means']
    test_scores_std = lc_df['test_scores_stds']
    fit_times_mean = lc_df['fit_times_means']
    fit_times_std = lc_df['fit_times_stds']

    # Plot learning curve
    axes[0].grid()
    axes[0].fill_between(train_sizes, train_scores_mean - train_scores_std,
                         train_scores_mean + train_scores_std, alpha=0.1,
                         color="r")
    axes[0].fill_between(train_sizes, test_scores_mean - test_scores_std,
                         test_scores_mean + test_scores_std, alpha=0.1,
                         color="g")
    axes[0].plot(train_sizes, train_scores_mean, 'o-', color="r",
                 label="Training score")
    axes[0].plot(train_sizes, test_scores_mean, 'o-', color="g",
                 label="Cross-validation score")
    axes[0].legend(loc="best")

    # Plot n_samples vs fit_times
    axes[1].grid()
    axes[1].plot(train_sizes, fit_times_mean, 'o-')
    axes[1].fill_between(train_sizes, fit_times_mean - fit_times_std,
                         fit_times_mean + fit_times_std, alpha=0.1)
    axes[1].set_xlabel("Training examples")
    axes[1].set_ylabel("fit_times")
    axes[1].set_title("Scalability of the model")

    # Plot fit_time vs score
    axes[2].grid()
    axes[2].plot(fit_times_mean, test_scores_mean, 'o-')
    axes[2].fill_between(fit_times_mean, test_scores_mean - test_scores_std,
                         test_scores_mean + test_scores_std, alpha=0.1)
    axes[2].set_xlabel("fit_times")
    axes[2].set_ylabel("Score")
    axes[2].set_title("Performance of the model")

    return fig



"""
__arus__ python package provides a computation framework to manage and process ubiquitous sensory data for activity recognition.

"""

__version__ = '1.1.18'

from . import accelerometer as accel
from . import cli
from . import developer as dev
from . import env
from . import extensions as ext
from . import mhealth_format as mh
from . import (operator, plugins, processor, segmentor, stream2, synchronizer,
               testing)
from .generator import *
from .moment import Moment
from .node import Node
from .scheduler import Scheduler
from .stream import Stream
from . import models
from .class_label import ClassSet
from . import feature_vector as fv
from . import dataset as ds
from . import spades_lab as slab



"""
synchronizer class that takes a set of chunks from multiple sources, sync and assemble and output them as a list.

Author: Qu Tang
Date: 02/07/2020
License: GNU v3
"""
from . import moment
from . import operator


class Synchronizer(operator.Operator):
    def __init__(self):
        super().__init__()
        self._buffer = {}
        self._num = 0
        pass

    def add_sources(self, n):
        self._num += n

    def remove_sources(self, n):
        self._num -= n

    def add_source(self):
        self._num += 1

    def remove_source(self):
        self._num -= 1
        self._num = max(0, self._num)

    def stop(self):
        super().stop()
        self.reset()

    def reset(self):
        self._buffer.clear()

    def run(self, values, src=None, context={}):
        """[summary]

        Arguments:
            values {[type]} -- [description]

        Keyword Arguments:
            src {[type]} -- [description] (default: {None})
            context {dict} -- It has to provide `data_id` as an indicator of the incoming data source, `start_time` and `stop_time` as indicators of the start and stop boundary of the data. (default: {{}})
        """
        new_context = self.merge_context(context)
        del new_context['data_id']
        st = new_context['start_time']
        et = new_context['stop_time']
        data_id = context['data_id']
        result, new_context = self._sync(values, st, et, data_id, new_context)
        if result is not None:
            self._result.put((result, new_context))

    def _sync(self, data, st, et, data_id, new_context):
        st = moment.Moment(st)
        et = moment.Moment(et)
        if st.to_unix_timestamp() not in self._buffer:
            self._buffer[st.to_unix_timestamp()] = {}
        # If source id exists, new data will overwrite old data
        self._buffer[st.to_unix_timestamp()][data_id] = data
        assembled = self._buffer[st.to_unix_timestamp()]
        if len(assembled.keys()) == self._num:
            # now the assemble is ready
            assembled, new_context = self._format_assembled(
                assembled, new_context)
            del self._buffer[st.to_unix_timestamp()]
            return assembled, new_context
        return None, new_context

    def _format_assembled(self, assembled, context):
        result = []
        data_ids = []
        for data_id, values in assembled.items():
            data_ids.append(data_id)
            result.append(values)
        new_context = {
            **context, "data_ids": data_ids
        }
        return result, new_context





import pandas as pd
import numpy as np
import pytest
from .. import class_label
from .. import mhealth_format as mh
from .. import extensions as ext
from .. import spades_lab as slab
from .. import extensions as ext


@pytest.fixture(scope='module')
def test_data(spades_lab_data):
    annotation_file = spades_lab_data['subjects']['SPADES_1']['annotations']['SPADESInLab'][0]
    data = pd.read_csv(annotation_file, parse_dates=[
                       0, 1, 2], infer_datetime_format=True)
    data = ext.pandas.segment_by_time(
        data, seg_st=data.iloc[0, 1], seg_et=data.iloc[0, 1] + pd.Timedelta(12.8, 's'), st_col=1, et_col=2)
    st = data.iloc[0, 1]
    et = data.iloc[-1, 2]
    return data, 'SPADES_1', st, et


@pytest.fixture(scope='module')
def test_data_multiwindows(spades_lab_data):
    annotation_file = spades_lab_data['subjects']['SPADES_1']['annotations']['SPADESInLab'][0]
    data = pd.read_csv(annotation_file, parse_dates=[
                       0, 1, 2], infer_datetime_format=True)
    data = ext.pandas.segment_by_time(
        data, seg_st=data.iloc[0, 1], seg_et=data.iloc[0, 1] + pd.Timedelta(12.8 * 5, 's'), st_col=1, et_col=2)
    st = data.iloc[0, 1]
    et = st + pd.Timedelta(12.8 * 5, 's')
    return data, 'SPADES_1', st, et


class TestClassSet:
    @ pytest.mark.parametrize('task_names', [['ACTIVITY_VALIDATED'], ['POSTURE_VALIDATED', 'INTENSITY']])
    def test_slab_class_set(self, test_data, task_names):

        raw_df, pid, st, et = test_data

        class_vector = slab.class_set(
            raw_df, st=st, et=et, task_names=task_names, pid=pid)

        np.testing.assert_array_equal(class_vector.columns,
                                      mh.FEATURE_SET_TIMESTAMP_COLS + task_names)

        assert class_vector.shape[0] == 1
        assert class_vector.notna().all(axis=None)

    @pytest.mark.parametrize('step_size', [None, 6.4])
    @pytest.mark.parametrize('start_time', [None, "2015-09-24 14:35:00.000"])
    @pytest.mark.parametrize('stop_time', [None, "2015-09-24 14:17:00.000"])
    @pytest.mark.parametrize('task_names', [['ACTIVITY_VALIDATED'], ['POSTURE_VALIDATED', 'INTENSITY']])
    def test_compute_offline(self, test_data_multiwindows, step_size, start_time, stop_time, task_names):
        raw_df, pid, st, et = test_data_multiwindows
        class_computer = class_label.ClassSet([raw_df], ['SPADESInLab'])
        class_computer.compute_offline(window_size=12.8,
                                       class_func=slab.class_set, task_names=task_names,
                                       step_size=step_size,
                                       start_time=start_time or st,
                                       stop_time=stop_time or et,
                                       pid=pid)
        class_set = class_computer.get_class_set()
        names = class_computer.get_task_names()

        if stop_time is not None:
            assert class_set is None
            np.testing.assert_array_equal(names, task_names)
        else:
            assert class_set.shape[1] == 3 + len(task_names)
            if step_size is None and start_time is None and stop_time is None:
                assert class_set.shape[0] == 5
                assert class_set.notna().all(axis=None)
            elif step_size is None and start_time is not None and stop_time is None:
                assert class_set.shape[0] == 6
                assert class_set.notna().all(axis=None)
            elif step_size == 6.4 and start_time is None and stop_time is None:
                assert class_set.shape[0] == 10
                assert class_set.notna().all(axis=None)
            elif step_size == 6.4 and start_time is not None and stop_time is None:
                assert class_set.shape[0] == 12
                assert class_set.notna().all(axis=None)
            elif stop_time is not None:
                assert class_set is None



try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    pass
from .. import extensions as ext
from loguru import logger
import shutil


class Reporter:
    def __init__(self, log_dir=None):
        self.log_dir = log_dir

    def reset(self):
        logger.info('Reset reporter log information...')
        shutil.rmtree(self.log_dir, ignore_errors=True)

    def report_cv(self, cv_result=None, cm_df=None, lc_df=None, tag_prefix='cv'):
        writer = SummaryWriter(log_dir=self.log_dir)
        if cv_result is not None:
            n_iter = cv_result.shape[0]
            logger.info('Report cv metrics...')
            for i in range(n_iter):
                test_metric_dict = {}
                train_metric_dict = {}
                time_dict = {}
                for j in range(len(cv_result.columns)):
                    if 'test' in cv_result.columns[j]:
                        test_metric_dict[cv_result.columns[j]
                                         ] = cv_result.iloc[i, j]
                    elif 'train' in cv_result.columns[j]:
                        train_metric_dict[cv_result.columns[j]
                                          ] = cv_result.iloc[i, j]
                    elif 'time' in cv_result.columns[j]:
                        time_dict[cv_result.columns[j]
                                  ] = cv_result.iloc[i, j]
                writer.add_scalars(f'{tag_prefix}/train', train_metric_dict, i)
                writer.add_scalars(f'{tag_prefix}/test', test_metric_dict, i)
                writer.add_scalars(f'{tag_prefix}/time', time_dict, i)
        if cm_df is not None:
            logger.info('Report cv confusion matrix')
            fig = ext.plotting.confusion_matrix(cm_df)
            writer.add_figure(
                f'{tag_prefix}/confusion_matrix', fig, close=True)
        if lc_df is not None:
            logger.info('Report cv learning curve')
            fig = ext.plotting.learning_curve(lc_df, ylim=[0, 1])
            writer.add_figure(
                f'{tag_prefix}/learning_curve', fig, close=True
            )
        writer.close()



from .. import mhealth_format as mh
from .. import moment
import pytest
import pandas as pd
from concurrent import futures
import os
import numpy as np
import datetime as dt
import sys


@pytest.fixture(scope='module')
def test_data(spades_lab_data):
    sensor_files = spades_lab_data['subjects']['SPADES_1']['sensors']['DW']
    data = pd.concat(
        map(lambda f: pd.read_csv(f, parse_dates=[0]), sensor_files), sort=False)
    return data


@pytest.fixture(scope='module', params=['sensors', 'annotations'])
def test_file(request, spades_lab_data):
    if request.param == 'sensors':
        return spades_lab_data['subjects']['SPADES_2']['sensors']['DW'][0]
    else:
        return spades_lab_data['subjects']['SPADES_2']['annotations']['SPADESInLab'][0]


@pytest.fixture(scope='module')
def filepath_test_cases():
    valid_filepaths = [
        'D:\\data\\spades_lab\\SPADES_7\\MasterSynced\\2015\\11\\19\\16',
        'D:\\data\\spades_lab\\SPADES_7\\MasterSynced\\2015\\11\\19\\16\\',
        'D:/data/spades_lab/SPADES_7/MasterSynced/2015/11/19/16/',
        'D:/data/spades_lab/SPADES_7/Derived/2015/11/19/16/',
        'D:/data/spades_lab/SPADES_7/Derived/AllSensors/2015/11/19/16/',
        'D:/data/spades_lab/SPADES_7/MasterSynced/2015/11/19/16',
        '''D:/data/spades_lab/SPADES_7/MasterSynced/2015/11/19/16/
            ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.
            2015-11-19-16-00-00-000-M0500.sensor.csv''',
        '''D:/data/spades_lab/SPADES_7/MasterSynced/2015/11/19/16/
            ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.
            2015-11-19-16-00-00-000-M0500.sensor.csv.gz'''
    ]

    invalid_filepaths = [
        'C:\\',
        'C:/',
        'D:/data/spades_lab/SPADES_7/MasterSynced/2015/16',
        'D:/data/spades_lab/SPADES_7/MasterSynced/20/16',
        'D:/data/spades_lab/SPADES_7/Mastenced/2015/16/17/21/'
    ]

    valid_flat_filepaths = [
        'D:\\data\\spades_lab\\SPADES_7\\MasterSynced\\asdf.csv',
        'D:\\data\\spades_lab\\SPADES_7\\MasterSynced\\adfa.csv',
        'D:/data/spades_lab/SPADES_7/MasterSynced/sdf.csv',
        'D:/data/spades_lab/SPADES_7/Derived/dfew.csv',
        'D:/data/spades_lab/SPADES_7/Derived/AllSensors/dfsd.csv',
        '''D:/data/spades_lab/SPADES_7/MasterSynced/ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv''',
        '''D:/data/spades_lab/SPADES_7/MasterSynced/ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv.gz''',
    ]

    invalid_flat_filepaths = [
        'C:\\',
        'C:/',
        'D:/data/spades_lab/SPADES_7/MasterSyn/',
        'D:/data/spades_lab/SPADES_7/MasterSynced/20/16',
        'D:/data/spades_lab/SPADES_7/Mastenced/2015/16/17/21/',
        '''D:/data/spades_lab/SPADES_7/MasterSynced/2015/11/19/16/ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv'''
    ]

    valid_filenames = [
        'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
        'SPADESInLab.diego-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv',
        'SPADESInLab.DIEGO-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv',
        'SPADESInLab.DIEGO-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv.gz'
    ]

    invalid_filenames = [
        'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.annotation.csv',
        'Actig?raphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
        'ActigraphGT9X-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
        'ActigraphGT9X-AccelerationCalibrated-0,1,2.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
        'ActigraphGT9X-AccelerationCalibrated-NA.tas1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
        'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-0-000-M0500.sensor.csv',
        'SPADESInLab-sdfsdf.diego-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv'
    ]

    valid_sensor_type = [
        'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
        'SPADESInLab.diego-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv',
        'SPADESInLab.DIEGO-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv',
        'SPADESInLab.DIEGO-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv.gz',
    ]

    return valid_filepaths, invalid_filepaths, valid_flat_filepaths, invalid_flat_filepaths, valid_filenames, invalid_filenames, valid_sensor_type


class TestCore:
    def test_get_session_start_time(self, spades_lab_data):
        start_time = mh.get_session_start_time(
            'SPADES_12', spades_lab_data['meta']['root'])
        assert start_time.strftime(
            '%Y-%m-%d-%H-%M-%S') == '2015-12-14-11-00-00'

    def test_get_session_span(self, spades_lab_data):
        session_st, session_et = mh.get_session_span(
            'SPADES_1', spades_lab_data['meta']['root'])
        assert session_st.strftime(
            '%Y-%m-%d-%H-%M-%S') == '2015-09-24-14-00-00'
        assert session_et.strftime(
            '%Y-%m-%d-%H-%M-%S') == '2015-09-24-17-00-00'

    def test_get_date_folders(self, spades_lab_data):
        date_folders = sorted(mh.get_date_folders(
            'SPADES_1', spades_lab_data['meta']['root']))
        test_dates = [
            '2015/09/24/14',
            '2015/09/24/15',
            '2015/09/24/16',
        ]
        for date_folder, test_date in zip(date_folders, test_dates):
            assert test_date in date_folder.replace(os.sep, '/')


class TestHelper:
    def test_is_mhealth_filepath(self, filepath_test_cases):
        for test_case in filepath_test_cases[0]:
            assert mh.is_mhealth_filepath(test_case)

        for test_case in filepath_test_cases[1]:
            assert not mh.is_mhealth_filepath(test_case)

    def test_is_mhealth_flat_filepath(self, filepath_test_cases):
        for test_case in filepath_test_cases[2]:
            assert mh.is_mhealth_flat_filepath(test_case)

        for test_case in filepath_test_cases[3]:
            assert not mh.is_mhealth_flat_filepath(test_case)

    def test_is_mhealth_filename(self, filepath_test_cases):

        for test_case in filepath_test_cases[4]:
            assert mh.is_mhealth_filename(test_case)

        for test_case in filepath_test_cases[5]:
            assert not mh.is_mhealth_filename(test_case)

    @pytest.mark.skipif(sys.platform == 'linux', reason="does not run on linux")
    def test_parse_pid_from_filepath_win(self):
        correct_test_cases = [
            'D:\\data\\spades_lab\\SPADES_7\\MasterSynced\\2015\\11\\19\\16',
            'D:\\data\\spades_lab\\SPADES_7\\MasterSynced\\2015\\11\\19\\16\\',
            'D:/data/spades_lab/SPADES_7/MasterSynced/2015/11/19/16/',
            'D:/data/spades_lab/SPADES_7/Derived/2015/11/19/16/',
            'D:/data/spades_lab/SPADES_7/Derived/AllSensors/2015/11/19/16/',
            'D:/data/spades_lab/SPADES_7/MasterSynced/2015/11/19/16',
            '''D:/data/spades_lab/SPADES_7/MasterSynced/2015/11/19/16/
            ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.
            2015-11-19-16-00-00-000-M0500.sensor.csv''',
            'D:\\data\\spades_lab\\SPADES_7\\MasterSynced\\asdf.csv',
            'D:\\data\\spades_lab\\SPADES_7\\MasterSynced\\adfa.csv',
            'D:/data/spades_lab/SPADES_7/MasterSynced/sdf.csv',
            'D:/data/spades_lab/SPADES_7/Derived/dfew.csv',
            'D:/data/spades_lab/SPADES_7/Derived/AllSensors/dfsd.csv',
            '''D:/data/spades_lab/SPADES_7/MasterSynced/ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv''',
            '''D:/data/spades_lab/SPADES_7/MasterSynced/ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv.gz'''
        ]

        incorrect_test_cases = [
            'C:\\',
            'C:/',
            'D:/data/spades_lab/SPADES_7/MasterSynced/2015/16',
            'D:/data/spades_lab/SPADES_7/MasterSynced/20/16',
            'D:/data/spades_lab/SPADES_7/Mastenced/2015/16/17/21/'
        ]

        for test_case in correct_test_cases:
            print(test_case)
            assert mh.parse_pid_from_filepath(test_case) == 'SPADES_7'

        for test_case in incorrect_test_cases:
            with pytest.raises(mh.ParseError):
                mh.parse_pid_from_filepath(test_case)

    def test_parse_sensor_type_from_filepath(self, filepath_test_cases):
        for test_case in filepath_test_cases[6]:
            print(test_case)
            assert mh.parse_sensor_type_from_filepath(
                test_case) == 'ActigraphGT9X' or mh.parse_sensor_type_from_filepath(test_case) == 'SPADESInLab'

    def test_parse_data_type_from_filepath(self):
        sensor_test_cases = [
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv.gz',
        ]

        for test_case in sensor_test_cases:
            assert mh.parse_data_type_from_filepath(
                test_case) == 'AccelerationCalibrated'

    def test_parse_version_code_from_filepath(self):
        sensor_test_cases = [
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv.gz',
        ]

        for test_case in sensor_test_cases:
            assert mh.parse_version_code_from_filepath(test_case) == 'NA'

    def test_parse_sensor_id_from_filepath(self):
        sensor_test_cases = [
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv.gz',
        ]

        for test_case in sensor_test_cases:
            assert mh.parse_sensor_id_from_filepath(
                test_case) == 'TAS1E23150152'

    def test_parse_file_type_from_filepath(self):
        sensor_test_cases = [
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv.gz',
        ]

        annotation_test_cases = [
            'SPADESInLab.diego-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv',
            'SPADESInLab.DIEGO-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv',
            'SPADESInLab.DIEGO-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv.gz',
        ]

        for test_case in sensor_test_cases:
            assert mh.parse_filetype_from_filepath(test_case) == 'sensor'

        for test_case in annotation_test_cases:
            assert mh.parse_filetype_from_filepath(test_case) == 'annotation'

    def test_parse_timestamp_from_filepath(self):
        sensor_test_cases = [
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv',
            'ActigraphGT9X-AccelerationCalibrated-NA.TAS1E23150152-AccelerationCalibrated.2015-11-19-16-00-00-000-M0500.sensor.csv.gz',
        ]

        annotation_test_cases = [
            'SPADESInLab.diego-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv',
            'SPADESInLab.DIEGO-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv',
            'SPADESInLab.DIEGO-SPADESInLab.2015-11-19-16-00-00-000-M0500.annotation.csv.gz',
        ]

        for test_case in sensor_test_cases:
            ts = mh.parse_timestamp_from_filepath(test_case, ignore_tz=True)
            ts_unix = moment.Moment(ts).to_unix_timestamp()
            assert dt.datetime.fromtimestamp(ts_unix) == ts

        for test_case in annotation_test_cases:
            ts = mh.parse_timestamp_from_filepath(test_case, ignore_tz=True)
            ts_unix = moment.Moment(ts).to_unix_timestamp()
            assert dt.datetime.fromtimestamp(ts_unix) == ts

    def test_parse_date_from_filepath(self, spades_lab_data):
        test_sensor_files = spades_lab_data['subjects']['SPADES_1']['sensors']['DW']
        test_dates = [
            '2015-09-24-14-00-00',
            '2015-09-24-15-00-00'
            '2015-09-24-16-00-00'
        ]
        for test_case, test_date in zip(test_sensor_files, test_dates):
            date = mh.parse_date_from_filepath(test_case)
            date.strftime('%Y-%m-%d-%H-%M-%S') == test_date

        test_annotation_files = spades_lab_data['subjects']['SPADES_1']['annotations']['SPADESInLab']
        test_dates = [
            '2015-09-24-14-00-00',
            '2015-09-24-15-00-00'
            '2015-09-24-16-00-00'
        ]
        for test_case, test_date in zip(test_annotation_files, test_dates):
            date = mh.parse_date_from_filepath(test_case)
            date.strftime('%Y-%m-%d-%H-%M-%S') == test_date

    def test_parse_subject_path_from_filepath(self, spades_lab_data):
        test_sensor_file = spades_lab_data['subjects']['SPADES_1']['sensors']['DW'][0]
        subject_path = mh.parse_subject_path_from_filepath(test_sensor_file)
        assert subject_path.endswith('SPADES_1')
        test_annotation_file = spades_lab_data['subjects']['SPADES_1']['annotations']['SPADESInLab'][0]
        subject_path = mh.parse_subject_path_from_filepath(
            test_annotation_file)
        assert subject_path.endswith('SPADES_1')

    def test_transform_class_category(self, spades_lab_data):
        class_category = spades_lab_data['meta']['class_category']
        input_category = 'FINEST_ACTIVITIES'
        output_category = 'MUSS_3_POSTURES'
        input_label = 'Lying on the back'
        assert mh.transform_class_category(
            input_label, class_category, input_category, output_category) == 'Lying'


class TestIO:
    @pytest.mark.parametrize('append', [False, True])
    @pytest.mark.parametrize('date_folders', [False, True])
    @pytest.mark.parametrize('hourly', [False, True])
    @pytest.mark.parametrize('block', [False, True])
    def test_mhealthfilewriter(self, test_data, hourly, date_folders, append, block, tmpdir):
        hour = test_data.iloc[0, 0].to_pydatetime().hour
        dataset_folder = tmpdir.realpath()
        pid = 'test'
        writer = mh.MhealthFileWriter(
            dataset_folder, pid, hourly=hourly, date_folders=date_folders)
        writer.set_for_sensor('test', 'test', 'test', 'NA')
        if append:
            first_hour_filename = writer._get_output_filename(test_data)
            first_hour_folder = writer._get_output_folder(test_data)
            os.makedirs(first_hour_folder, exist_ok=True)
            existing_filepath = os.path.join(
                first_hour_folder, first_hour_filename)
            with open(existing_filepath, 'w'):
                pass
            assert os.path.exists(existing_filepath)

        results = writer.write_csv(test_data, append=append, block=block)
        if block:
            output_paths = results
            existings = [os.path.exists(f) for f in output_paths]
            np.testing.assert_array_equal(existings, True)
        else:
            done, _ = futures.wait(results)
            output_paths = [f.result() for f in done]
            existings = [os.path.exists(f) for f in output_paths]
            np.testing.assert_array_equal(existings, True)
        if not hourly:
            assert len(output_paths) == 1
        else:
            assert len(output_paths) == 3
        parent_folder_names = [os.path.basename(
            os.path.dirname(f)) for f in output_paths]
        if date_folders and hourly:
            np.testing.assert_array_equal(
                sorted(parent_folder_names), np.arange(start=hour, step=1, stop=17).astype(str))
        elif date_folders and not hourly:
            np.testing.assert_array_equal(
                sorted(parent_folder_names), str(hour))
        else:
            np.testing.assert_array_equal(
                parent_folder_names, [mh.MASTER_FOLDER]*len(output_paths))

    @pytest.mark.parametrize('chunksize', [None, 10])
    def test_mhealthfilereader(self, test_file, chunksize):
        reader = mh.MhealthFileReader(test_file)
        file_type = mh.parse_filetype_from_filepath(test_file)
        if file_type == 'sensor':
            datetime_cols = [0]
        else:
            datetime_cols = [0, 1, 2]
        reader.read_csv(chunksize=chunksize, datetime_cols=datetime_cols)
        results = []
        for data in reader.get_data():
            if chunksize is None:
                assert data.shape[0] > 10
            else:
                assert data.shape[0] == 10
            assert data.columns[0] == mh.TIMESTAMP_COL
            if file_type == 'annotation':
                np.testing.assert_array_equal(
                    data.columns[:3], mh.FEATURE_SET_TIMESTAMP_COLS)
            results.append(data)
            if len(results) == 10:
                break
        if chunksize is None:
            assert len(results) == 1
        else:
            assert len(results) > 5



import enum

import PySimpleGUI as sg

import app_state as app
import backend
import base
import components as comp
import task_window
import stream_window
from arus.models.muss import Strategy


class Event(enum.Enum):
    TEST_ORIGIN_MODEL = enum.auto(),
    ORIGIN_LABELS_CHANGED = enum.auto(),
    NOT_ENOUGH_ORIGIN_LABELS = enum.auto(),
    ENOUGH_ORIGIN_LABELS = enum.auto(),
    TRAIN_ORIGIN_MODEL = enum.auto(),
    ORIGIN_MODEL_READY = enum.auto(),
    VALIDATE_ORIGIN_MODEL = enum.auto(),
    NEW_ACTIVITY_NAME_CHANGED = enum.auto(),
    ADD_NEW_ACTIVITY = enum.auto(),
    DATA_COLLECTION_LABELS_CHANGED = enum.auto(),
    NOT_ENOUGH_DATA_COLLECTION_LABELS = enum.auto(),
    ENOUGH_DATA_COLLECTION_LABELS = enum.auto(),
    COLLECT_NEW_DATA = enum.auto(),
    NEW_DATASET_READY = enum.auto(),
    CHECK_NEW_DATA = enum.auto(),
    NEW_LABELS_CHANGED = enum.auto(),
    NOT_ENOUGH_NEW_LABELS = enum.auto(),
    ENOUGH_NEW_LABELS = enum.auto(),
    TRAIN_NEW_MODEL = enum.auto(),
    NEW_MODEL_READY = enum.auto(),
    TEST_NEW_MODEL = enum.auto(),
    VALIDATE_NEW_MODEL = enum.auto(),
    ORIGIN_VALIDATE_RESULT_READY = enum.auto(),
    NEW_VALIDATE_RESULT_READY = enum.auto()


class Dashboard(base.BaseWindow):
    def __init__(self, title):
        super().__init__(title, 25)

    def init_states(self):
        state = app.AppState.getInstance()
        origin_label_candidates = backend.get_class_label_candidates(
            state.origin_dataset)
        if len(state.data_collection_label_candidates) == 0:
            state.data_collection_label_candidates = origin_label_candidates
        if len(state.origin_label_candidates) == 0:
            state.origin_label_candidates = origin_label_candidates
        return state

    def init_views(self):
        self._window_train_origin = None
        self._window_data_collection = None
        self._window_train_new = None
        self._window_validate_origin = None
        self._window_validate_new = None
        origin_model_column = self.init_origin_model_view()
        data_collection_column = self.init_data_collection_view()
        new_model_column = self.init_new_model_view()

        layout = [
            [
                origin_model_column,
                data_collection_column,
                new_model_column
            ]
        ]
        dashboard = sg.Window(
            self._title,
            layout=layout,
            finalize=True,
            resizable=True
        )
        return dashboard

    def _update_states_and_events(self, event, values):
        if self._state.origin_labels != values[Event.ORIGIN_LABELS_CHANGED]:
            self._state.origin_labels = values[Event.ORIGIN_LABELS_CHANGED][:]
            if len(self._state.origin_labels) >= 2:
                self._events.put(Event.ENOUGH_ORIGIN_LABELS)
            else:
                self._events.put(Event.NOT_ENOUGH_ORIGIN_LABELS)

        if self._state.new_labels != values[Event.NEW_LABELS_CHANGED]:
            self._state.new_labels = values[Event.NEW_LABELS_CHANGED][:]
            if len(self._state.new_labels) >= 1:
                self._events.put(Event.ENOUGH_NEW_LABELS)
            else:
                self._events.put(Event.NOT_ENOUGH_NEW_LABELS)

        if self._state.data_collection_labels != values[Event.DATA_COLLECTION_LABELS_CHANGED]:
            self._state.data_collection_labels = values[
                Event.DATA_COLLECTION_LABELS_CHANGED][:]
            if len(self._state.data_collection_labels) >= 1:
                self._events.put(Event.ENOUGH_DATA_COLLECTION_LABELS)
            else:
                self._events.put(Event.NOT_ENOUGH_DATA_COLLECTION_LABELS)

        if self._state.new_activity_name != values[Event.NEW_ACTIVITY_NAME_CHANGED]:
            self._state.new_activity_name = values[Event.NEW_ACTIVITY_NAME_CHANGED]

        if event == Event.ADD_NEW_ACTIVITY:
            if self._state.new_activity_name not in self._state.data_collection_label_candidates:
                self._state.data_collection_label_candidates = [
                    self._state.new_activity_name] + self._state.data_collection_label_candidates

        if self._window_train_origin is not None:
            origin_model = self._window_train_origin.get_result()
            if origin_model is not None:
                self._state.origin_model = origin_model
                self._window_train_origin = None
                self._events.put(Event.ORIGIN_MODEL_READY)

        if self._window_data_collection is not None:
            new_dataset = self._window_data_collection.get_result()
            if new_dataset is not None:
                self._state.new_dataset = new_dataset
                self._window_data_collection = None
                self._events.put(Event.NEW_DATASET_READY)

        if self._window_train_new is not None:
            new_model = self._window_train_new.get_result()
            if new_model is not None:
                self._state.new_model = new_model
                self._window_train_new = None
                self._events.put(Event.NEW_MODEL_READY)

        if self._window_validate_origin is not None:
            origin_validate_results = self._window_validate_origin.get_result()
            if origin_validate_results is not None:
                self._state.origin_validate_results = origin_validate_results
                self._window_validate_origin = None
                self._events.put(Event.ORIGIN_VALIDATE_RESULT_READY)

        if self._window_validate_new is not None:
            new_validate_results = self._window_validate_new.get_result()
            if new_validate_results is not None:
                self._state.new_validate_results = new_validate_results
                self._window_validate_new = None
                self._events.put(Event.NEW_VALIDATE_RESULT_READY)

    def _dispatch_events(self, event):
        if event == Event.ENOUGH_ORIGIN_LABELS:
            self.enable_train_origin_model()
        elif event == Event.NOT_ENOUGH_ORIGIN_LABELS:
            self.disable_train_origin_model()
        elif event == Event.ENOUGH_DATA_COLLECTION_LABELS:
            self.enable_collect_data()
        elif event == Event.NOT_ENOUGH_DATA_COLLECTION_LABELS:
            self.disable_collect_data()
        elif event == Event.ENOUGH_NEW_LABELS:
            self.enable_train_new_model()
        elif event == Event.NOT_ENOUGH_NEW_LABELS:
            self.disable_train_new_model()
        elif event == Event.TRAIN_ORIGIN_MODEL:
            self.open_task_window(event)
        elif event == Event.ORIGIN_MODEL_READY:
            self.enable_test_origin_model()
            self.display_model_summary(event)
        elif event == Event.VALIDATE_ORIGIN_MODEL:
            self.open_task_window(event)
        elif event == Event.TEST_ORIGIN_MODEL:
            self.open_stream_window(event)
        elif event == Event.COLLECT_NEW_DATA:
            self.open_stream_window(event)
        elif event == Event.NEW_DATASET_READY:
            self.enable_check_data_button()
            self.display_dataset_summary(event)
            self.update_new_label_list()
        elif event == Event.ADD_NEW_ACTIVITY:
            self.update_data_collection_label_list()
            self.reset_new_activity_input()
        elif event == Event.TRAIN_NEW_MODEL:
            self.open_task_window(event)
        elif event == Event.VALIDATE_NEW_MODEL:
            self.open_task_window(event)
        elif event == Event.NEW_MODEL_READY:
            self.enable_test_new_model()
            self.display_model_summary(event)
        elif event == Event.TEST_NEW_MODEL:
            self.open_stream_window(event)

    def init_origin_model_view(self):
        heading = 'Step 1 - Use origin data'
        description = "Hold 'Ctrl' to select multiple classes you would like to train for the origin model."
        button_texts = ['Train model', 'Validate model', 'Test model']
        model_info = backend.get_model_summary(
            model=self._state.origin_model)

        header_row = [
            comp.heading(
                heading,
                fixed_column_width=self._column_width
            )
        ]

        description_row = [
            comp.text(
                description,
                fixed_column_width=self._column_width
            )
        ]

        label_selection_row = [
            comp.selection_list(
                items=self._state.origin_label_candidates,
                default_selections=self._state.origin_labels,
                fixed_column_width=self._column_width,
                mode='multiple',
                key=Event.ORIGIN_LABELS_CHANGED,
                rows=20
            )
        ]

        self._train_button_origin_model = comp.control_button(
            text=button_texts[0],
            fixed_column_width=self._column_width,
            disabled=len(self._state.origin_labels) == 0,
            key=Event.TRAIN_ORIGIN_MODEL
        )
        train_button_row = [
            self._train_button_origin_model
        ]

        self._validate_button_origin_model = comp.control_button(
            text=button_texts[1],
            fixed_column_width=self._column_width,
            disabled=self._state.origin_model is None, key=Event.VALIDATE_ORIGIN_MODEL
        )
        validate_button_row = [
            self._validate_button_origin_model
        ]

        self._test_button_origin_model = comp.control_button(
            text=button_texts[2],
            fixed_column_width=self._column_width,
            disabled=self._state.origin_model is None,
            key=Event.TEST_ORIGIN_MODEL
        )

        test_button_row = [
            self._test_button_origin_model
        ]

        self._text_origin_model = comp.text(
            text=model_info,
            fixed_column_width=self._column_width
        )

        model_info_row = [
            self._text_origin_model
        ]

        layout = [
            header_row,
            description_row,
            label_selection_row,
            train_button_row,
            validate_button_row,
            test_button_row,
            model_info_row
        ]

        return sg.Column(layout=layout, scrollable=False, element_justification='center')

    def init_data_collection_view(self):
        heading = 'Step 2 - Collect new data'
        description = "Type a new activity class or select an activity class from the dropdown list that you want to collect data for. Hold 'Ctrl' to select multiple classes you would like to collect data for."

        confirm_button_text = 'Add'

        data_info = backend.get_dataset_summary(
            dataset=self._state.new_dataset)

        button_texts = ['Collect data', 'Check data']

        header_row = [
            comp.heading(
                text=heading,
                fixed_column_width=self._column_width
            )
        ]

        description_row = [
            comp.text(
                text=description,
                fixed_column_width=self._column_width
            )
        ]

        self._input_new_activity = comp.TextInput(
            confirm_button_text=confirm_button_text,
            candidate_items=[],
            fixed_column_width=self._column_width,
            confirm_button_key=Event.ADD_NEW_ACTIVITY,
            text_input_key=Event.NEW_ACTIVITY_NAME_CHANGED,
        )
        input_row = self._input_new_activity.get_component_as_row()

        self._data_collection_label_list = comp.selection_list(
            items=self._state.data_collection_label_candidates,
            mode='multiple',
            fixed_column_width=self._column_width,
            default_selections=self._state.data_collection_labels,
            key=Event.DATA_COLLECTION_LABELS_CHANGED,
            rows=17
        )
        label_selection_row = [
            self._data_collection_label_list
        ]

        self._collect_button_new_data = comp.control_button(
            text=button_texts[0],
            key=Event.COLLECT_NEW_DATA,
            fixed_column_width=self._column_width,
            disabled=self._state.data_collection_labels is None
        )
        collect_button_row = [
            self._collect_button_new_data
        ]

        self._check_button_new_data = comp.control_button(
            text=button_texts[1],
            key=Event.CHECK_NEW_DATA,
            fixed_column_width=self._column_width,
            disabled=self._state.new_dataset is None
        )
        check_button_row = [
            self._check_button_new_data
        ]

        self._text_new_data = comp.text(
            text=data_info,
            fixed_column_width=self._column_width
        )
        data_info_row = [
            self._text_new_data
        ]

        layout = [
            header_row,
            description_row,
            input_row,
            label_selection_row,
            collect_button_row,
            check_button_row,
            data_info_row
        ]

        return sg.Column(layout=layout, scrollable=False, element_justification='center')

    def init_new_model_view(self):
        heading = 'Step 3 - Use new data'
        description = "Hold 'Ctrl' to select multiple classes you would like to use the new data to train a new model."
        button_texts = ['Train model', 'Validate model', 'Test model']
        model_info = backend.get_model_summary(self._state.new_model)
        new_data = self._state.new_dataset
        if new_data is None:
            new_label_candidates = []
        else:
            new_label_candidates = new_data['GT_LABEL'].unique().tolist()

        header_row = [
            comp.heading(
                heading,
                fixed_column_width=self._column_width
            )
        ]

        description_row = [
            comp.text(
                description,
                fixed_column_width=self._column_width
            )
        ]

        self._new_label_list = comp.selection_list(
            items=new_label_candidates,
            default_selections=self._state.new_labels,
            fixed_column_width=self._column_width,
            mode='multiple',
            key=Event.NEW_LABELS_CHANGED,
            rows=20
        )
        label_selection_row = [
            self._new_label_list
        ]

        self._train_button_new_model = comp.control_button(
            text=button_texts[0],
            fixed_column_width=self._column_width,
            disabled=self._state.new_labels is None, key=Event.TRAIN_NEW_MODEL
        )
        train_button_row = [
            self._train_button_new_model
        ]

        self._validate_button_new_model = comp.control_button(
            text=button_texts[1],
            fixed_column_width=self._column_width,
            disabled=self._state.new_model is None,
            key=Event.VALIDATE_NEW_MODEL
        )
        validate_button_row = [
            self._validate_button_new_model
        ]

        self._test_button_new_model = comp.control_button(
            text=button_texts[2],
            fixed_column_width=self._column_width,
            disabled=self._state.new_model is None,
            key=Event.TEST_NEW_MODEL
        )

        test_button_row = [
            self._test_button_new_model
        ]

        self._text_new_model = comp.text(
            text=model_info,
            fixed_column_width=self._column_width
        )

        model_info_row = [
            self._text_new_model
        ]

        layout = [
            header_row,
            description_row,
            label_selection_row,
            train_button_row,
            validate_button_row,
            test_button_row,
            model_info_row
        ]

        return sg.Column(layout=layout, scrollable=False, element_justification='center')

    def open_task_window(self, event):
        if event == Event.TRAIN_ORIGIN_MODEL:
            self._window_train_origin = task_window.TaskWindow(
                "Train origin model")
            self._window_train_origin.start()
        elif event == Event.TRAIN_NEW_MODEL:
            train_strategies = [
                Strategy.REPLACE_ORIGIN,
                Strategy.COMBINE_ORIGIN,
                Strategy.USE_NEW_ONLY
            ]
            self._window_train_new = task_window.TaskWindow(
                'Train new model',
                strategies=train_strategies
            )
            self._window_train_new.start()
        elif event == Event.VALIDATE_ORIGIN_MODEL:
            self._window_validate_origin = task_window.TaskWindow(
                "Validate origin model", mode=task_window.Mode.LOSO)
            self._window_validate_origin.start()
        elif event == Event.VALIDATE_NEW_MODEL:
            strategies = [
                Strategy.REPLACE_ORIGIN,
                Strategy.COMBINE_ORIGIN
            ]
            self._window_validate_new = task_window.TaskWindow(
                "Validate new model", strategies=strategies, mode=task_window.Mode.LOSO)
            self._window_validate_new.start()

    def display_model_summary(self, event):
        if event == Event.ORIGIN_MODEL_READY:
            summary = backend.get_model_summary(self._state.origin_model)
            self._text_origin_model.update(summary)
        elif event == Event.NEW_MODEL_READY:
            summary = backend.get_model_summary(self._state.new_model)
            self._text_new_model.update(summary)

    def display_dataset_summary(self, event):
        if event == Event.NEW_DATASET_READY:
            summary = backend.get_dataset_summary(self._state.new_dataset)
            self._text_new_data.update(summary)

    def open_stream_window(self, event):
        if event == Event.TEST_ORIGIN_MODEL:
            panel = stream_window.StreamWindow(
                "Test origin model",
                model=self._state.origin_model,
                support_modes=[
                    backend.PROCESSOR_MODE.TEST_ONLY,
                    backend.PROCESSOR_MODE.TEST_AND_SAVE
                ]
            )
            panel.start()
        elif event == Event.TEST_NEW_MODEL:
            panel = stream_window.StreamWindow(
                "Test new model",
                model=self._state.new_model,
                support_modes=[
                    backend.PROCESSOR_MODE.TEST_ONLY,
                    backend.PROCESSOR_MODE.TEST_AND_SAVE
                ]
            )
            panel.start()
        elif event == Event.COLLECT_NEW_DATA:
            self._window_data_collection = stream_window.StreamWindow(
                'Collect new data',
                model=self._state.origin_model,
                labels=self._state.data_collection_labels,
                support_modes=[
                    backend.PROCESSOR_MODE.TEST_AND_COLLECT,
                    backend.PROCESSOR_MODE.COLLECT_ONLY,
                    backend.PROCESSOR_MODE.ACTIVE_COLLECT
                ]
            )
            self._window_data_collection.start()

    def enable_train_origin_model(self):
        self._train_button_origin_model.update(disabled=False)
        self._validate_button_origin_model.update(disabled=False)

    def enable_train_new_model(self):
        self._train_button_new_model.update(disabled=False)
        self._validate_button_new_model.update(disabled=False)

    def enable_test_origin_model(self):
        self._test_button_origin_model.update(disabled=False)

    def enable_test_new_model(self):
        self._test_button_new_model.update(disabled=False)

    def enable_collect_data(self):
        self._collect_button_new_data.update(disabled=False)

    def disable_train_origin_model(self):
        self._train_button_origin_model.update(disabled=True)
        self._validate_button_origin_model.update(disabled=True)

    def disable_train_new_model(self):
        self._train_button_new_model.update(disabled=True)
        self._validate_button_new_model.update(disabled=True)

    def disable_test_origin_model(self):
        self._test_button_origin_model.update(disabled=False)

    def disable_test_new_model(self):
        self._test_button_new_model.update(disabled=False)

    def disable_collect_data(self):
        self._collect_button_new_data.update(disabled=True)

    def disable_check_data_button(self):
        self._check_button_new_data.update(disabled=True)

    def enable_check_data_button(self):
        self._check_button_new_data.update(disabled=False)

    def update_data_collection_label_list(self):
        self._data_collection_label_list.update(
            self._state.data_collection_label_candidates)

    def reset_new_activity_input(self):
        self._input_new_activity.update_input(value="")

    def update_new_label_list(self):
        new_label_candidates = self._state.new_dataset['GT_LABEL'].unique(
        ).tolist()
        self._new_label_list.update(new_label_candidates)



import numpy as np
from .. import generator, segmentor, stream, developer
import logging
import pandas as pd
from .. import mhealth_format as mh
from ..models import muss as arus_muss
from ..models import nn
from ..core import pipeline as arus_pipeline
import os


def process_mehealth_dataset(dataset_dict, approach='muss', **kwargs):
    kwargs = _parse_kwargs(approach, kwargs)

    results = []

    for pid in dataset_dict['subjects'].keys():
        logging.info('Start processing {}'.format(pid))
        if approach == 'muss':
            processed = _process_muss(
                dataset_dict, pid, sr=kwargs['sr'], window_size=kwargs['window_size'])
            results.append(processed)
        elif approach == 'nn':
            processed = _process_nn(dataset_dict, pid, sr=kwargs['sr'])
            if processed is not None:
                processed.reset_index(drop=False, inplace=True)
                results.append(processed)
        else:
            raise NotImplementedError('Only "muss" approach is implemented.')

    processed_dataset = pd.concat(results, axis=0, sort=False)
    if approach == 'muss':
        processed_dataset.sort_values(
            by=[mh.FEATURE_SET_PID_COL, mh.FEATURE_SET_PLACEMENT_COL] + mh.FEATURE_SET_TIMESTAMP_COLS, inplace=True)
    return processed_dataset


def _parse_kwargs(approach, kwargs):
    if approach == 'muss' and 'window_size' not in kwargs:
        kwargs['window_size'] = 12.8
    elif approach == 'muss' and 'window_size' in kwargs:
        pass
    elif approach == 'nn':
        kwargs['window_size'] = 3600
    else:
        raise NotImplementedError('You must provide a valid window size')

    if 'sr' not in kwargs:
        raise NotImplementedError('You must provide a valid sampling rate')

    developer.logging_dict(kwargs)
    logging.info('sr: {}'.format(kwargs['sr']))
    logging.info('window size: {}'.format(kwargs['window_size']))
    return kwargs


def _process_muss(dataset_dict, pid, sr, window_size=12.8):
    dataset_path = dataset_dict['meta']['root']
    start_time = mh.get_session_start_time(pid, dataset_path)
    streams, streams_kwargs = _prepare_mhealth_streams(
        dataset_dict, pid, window_size, sr)
    streams_kwargs['dataset_name'] = dataset_dict['meta']['name']
    streams_kwargs['pid'] = pid
    pipeline = arus_muss.MUSSModel.get_mhealth_dataset_pipeline(
        *streams, name='{}-pipeline'.format(pid), scheduler='processes', max_processes=os.cpu_count() - 4, **streams_kwargs)  # os.cpu_count() - 4
    pipeline.start(start_time=start_time)
    processed = _prepare_pipeline_output(pipeline, pid)
    return processed


def _prepare_mhealth_streams(dataset_dict, pid, window_size, sr, use_annotations=True):
    streams = []
    subject_data_dict = dataset_dict['subjects']
    streams_kwargs = {}
    # sensor streams
    for p in subject_data_dict[pid]['sensors'].keys():
        stream_name = p
        gr = generator.MhealthSensorFileGenerator(
            *subject_data_dict[pid]['sensors'][p])
        seg = segmentor.SlidingWindowSegmentor(window_size=window_size)
        pid_sid_stream = stream.Stream(gr, seg, name=stream_name)
        streams.append(pid_sid_stream)
        streams_kwargs[stream_name] = {
            'sr': sr
        }

    if use_annotations:
        # annotation streams
        for a in subject_data_dict[pid]['annotations'].keys():
            stream_name = a
            gr = generator.MhealthAnnotationFileGenerator(
                *subject_data_dict[pid]['annotations'][a])
            seg = segmentor.SlidingWindowSegmentor(
                window_size=window_size, st_col=1, et_col=2)
            annotation_stream = stream.Stream(gr, seg, name=stream_name)
            streams.append(annotation_stream)

    return streams, streams_kwargs


def _prepare_pipeline_output(pipeline, pid):
    processed = None
    for df, st, et, prev_st, prev_et, name in pipeline.get_iterator():
        if df.empty:
            continue
        if processed is not None:
            processed = pd.concat(
                (processed, df), sort=False, axis=0)
        else:
            processed = df
    logging.info('Pipeline {} has completed.'.format(pid))
    pipeline.stop()
    if processed is not None:
        processed[mh.FEATURE_SET_PID_COL] = pid
    return processed


def _process_nn(dataset_dict, pid, sr):
    window_size = 3600
    dataset_path = dataset_dict['meta']['root']
    start_time = mh.get_session_start_time(
        pid, dataset_path, round_to='minute')
    streams, streams_kwargs = _prepare_mhealth_streams(
        dataset_dict, pid, window_size, sr, use_annotations=False)
    streams_kwargs['dataset_name'] = dataset_dict['meta']['name']
    streams_kwargs['pid'] = pid
    pipeline = arus_pipeline.Pipeline(
        max_processes=os.cpu_count() - 4, scheduler='processes', name='process-nn-pipeline')
    for stream in streams:
        pipeline.add_stream(stream)
    pipeline.set_processor(nn.preprocess_processor, **streams_kwargs)
    pipeline.start(start_time=start_time)
    pid = pid.split('_')[1]
    processed = _prepare_pipeline_output(pipeline, pid)
    return processed



from joblib import Parallel, delayed
from collections.abc import Iterable
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import StandardScaler
from scipy.linalg import pinv2
import numpy as np
from scipy.special import softmax
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels
from sklearn.metrics import accuracy_score
from sklearn.preprocessing import LabelEncoder


def _val_predict(clf, X_train, y_train, X_test, y_test, fit_params=None, method='predict'):
    fit_params = fit_params or {}
    clf.fit(X_train, y_train, **fit_params)
    predict = getattr(clf, method)
    y_pred = predict(X_test)
    return y_test, y_pred


def cross_val_predict(clf, X, y, groups=None, cv=None, n_jobs=None, fit_params=None, method='predict'):
    if isinstance(cv, Iterable):
        cv = cv
    elif hasattr(cv, "split") and hasattr(cv, "get_n_splits"):
        cv = cv.split(X, y, groups=groups)
    with Parallel(n_jobs=n_jobs) as parallel:
        results = parallel(delayed(_val_predict)(
            clf=clf,
            X_train=X[train_idx, :],
            y_train=y[train_idx],
            X_test=X[test_idx, :],
            y_test=y[test_idx],
            fit_params=fit_params,
            method=method) for train_idx, test_idx in cv)
    preds = []
    tests = []
    for test, pred in results:
        tests += test.tolist()
        preds += pred.tolist()
    return tests, preds


class ELM(BaseEstimator):
    def __init__(self, hidden_layer_sizes=(10000,), random_state=None):
        self.hidden_layer_sizes = hidden_layer_sizes
        self.random_state = random_state
        if random_state is None:
            self.random_state = np.random.randint(100)

    def _init_hidden_layers(self, input_size):
        np.random.seed(self.random_state)
        hidden_layer_sizes = self.hidden_layer_sizes
        self._input_layer_weights = np.random.normal(
            size=[input_size, hidden_layer_sizes[0]])
        self._input_layer_biases = np.random.normal(
            size=[hidden_layer_sizes[0]])

    def _activate(self, fn='relu'):
        if fn == 'relu':
            return lambda x: np.maximum(x, 0, x)

    def _feedforward(self, X):
        out1 = np.dot(X, self._input_layer_weights)
        out1 = out1 + self._input_layer_biases
        out2 = self._activate(fn='relu')(out1)
        return out2

    def _encode_y(self, y, labels=None, from_fit=True):
        if from_fit:
            if labels is None:
                labels = 'auto'
            else:
                labels = [labels]
            self._y_encoder = OneHotEncoder(categories=labels)
            return self._y_encoder.fit_transform(np.expand_dims(y, axis=1)).toarray().astype(int)
        else:
            return self._y_encoder.transform(y).toarray().astype(int)

    def _get_labels(self, y):
        self._labels = np.unique(y)

    def fit(self, X, y, labels=None):
        X = np.real(X).astype(float)
        X, y = check_X_y(X, y)
        self.classes_ = unique_labels(y)
        self.n_features_in_ = X.shape[1]

        self._init_hidden_layers(X.shape[1])
        encoded_y = self._encode_y(y, labels=labels, from_fit=True)
        out = self._feedforward(X)
        self._output_layer_weights = np.dot(pinv2(out), encoded_y)
        return self

    def _predict_scores(self, X):
        check_is_fitted(self)
        X = np.real(X).astype(float)
        X = check_array(X)
        out = self._feedforward(X)
        out = np.dot(out, self._output_layer_weights)
        return out

    def predict(self, X):
        scores = self._predict_scores(X)
        encoded_scores = np.zeros(scores.shape)
        for i in range(scores.shape[0]):
            encoded_scores[i] = (scores[i] == np.max(scores[i])).astype(int)
        pred_y = self._y_encoder.inverse_transform(encoded_scores)
        return pred_y

    def predict_proba(self, X):
        scores = self._predict_scores(X)
        pred_probs = softmax(scores, axis=1)
        return pred_probs

    def score(self, X, y):
        y_pred = self.predict(X)
        return accuracy_score(y, y_pred)


class LabelEncoderExt(object):
    def __init__(self):
        """
        It differs from LabelEncoder by handling new classes and providing a value for it [Unknown]
        Unknown will be added in fit and transform will take care of new item. It gives unknown class id
        """
        self.label_encoder = LabelEncoder()
        # self.classes_ = self.label_encoder.classes_

    def fit(self, data_list):
        """
        This will fit the encoder for all the unique values and introduce unknown value
        :param data_list: A list of string
        :return: self
        """
        self.label_encoder = self.label_encoder.fit(
            list(data_list) + ['Unknown'])
        self.classes_ = self.label_encoder.classes_

        return self

    def transform(self, data_list):
        """
        This will transform the data_list to id list where the new values get assigned to Unknown class
        :param data_list:
        :return:
        """
        new_data_list = list(data_list)
        for unique_item in np.unique(data_list):
            if unique_item not in self.label_encoder.classes_:
                new_data_list = ['Unknown' if x ==
                                 unique_item else x for x in new_data_list]

        return self.label_encoder.transform(new_data_list)

    def inverse_transform(self, id_list):
        unknown_id = self.label_encoder.transform(['Unknown'])[0]
        new_id_list = list(id_list)
        for unique_item in np.unique(id_list):
            if unique_item == -1:
                new_id_list = [unknown_id if x ==
                               unique_item else x for x in new_id_list]

        return self.label_encoder.inverse_transform(new_id_list)




import os
import app_state as app
try:
    import PySimpleGUI as sg
    import backend
    import session_panel
    import dashboard
except ImportError as e:
    msg = (
        "Arus demo requirements are not installed.\n\n"
        "Please install the demo extra packages as follows:\n\n"
        "  pip install arus[demo]\n\n"
        "  poetry add arus  --extras demo"
    )
    raise ImportError(str(e) + "\n\n" + msg)
import multiprocessing
import sys
import os


def start_app():
    multiprocessing.freeze_support()
    sg.theme('Reddit')
    os.chdir(os.path.dirname(os.path.realpath(__file__)))
    session_panel.SessionSelectionPanel().start()


if __name__ == "__main__":
    start_app()
